from .basic import Chatbot


class BoseBot(Chatbot):
    """Represents an agent in Navigator.
        https://eminent-kiwi-trusty.ngrok-free.app/api/bose/messages
        Each agent has a name, a role, a goal, a backstory,
        and an optional language model (llm).
    """
    name: str = 'BoseBot'
    company: str = 'T-ROC Global'
    company_website: str = 'https://www.trocglobal.com'
    contact_information = 'communications@trocglobal.com'
    contact_form = 'https://bose.trocdigital.io/bose/bose_ticketing_system'
    role: str = 'Bose Sound Systems Expert Technician and Consultant.'
    goal = 'Bring useful information to Bose Technicians and Consultants.'
    specialty_area = 'Bose endcap displays that are used to showcase Bose products'



from langchain_google_genai import (
    GoogleGenerativeAI,
    ChatGoogleGenerativeAI,
    GoogleGenerativeAIEmbeddings
)
from navconfig import config
from .abstract import AbstractLLM


class GoogleGenAI(AbstractLLM):
    """GoogleGenAI.
        Using Google Generative AI models with Google Cloud AI Platform.
    """
    model: str = "gemini-pro"
    supported_models: list = [
        "models/text-bison-001",
        "models/chat-bison-001",
        "gemini-pro"
    ]

    def __init__(self, *args, **kwargs):
        self.model_type = kwargs.get("model_type", "chat")
        super().__init__(*args, **kwargs)
        self._api_key = kwargs.pop('api_key', config.get('GOOGLE_API_KEY'))
        if self.model_type == 'chat':
            base_llm = ChatGoogleGenerativeAI
        else:
            base_llm = GoogleGenerativeAI
        self._llm = base_llm(
            model=self.model,
            api_key=self._api_key,
            temperature=self.temperature,
            **self.args
        )
        embed_model = kwargs.get("embed_model", "models/embedding-001")
        self._embed = GoogleGenerativeAIEmbeddings(
            model=embed_model,
            google_api_key=self._api_key,
            temperature=self.temperature,
            top_p=self.top_p,
            top_k=self.top_k,
        )



import asyncio
from pymilvus import connections
from parrot.chatbots.basic import Chatbot
from parrot.llms.groq import GroqLLM



_HOST = '127.0.0.1'
_PORT = '19530'

def create_connection():
    print(f"\nCreate connection...")
    connections.connect('test', host=_HOST, port=_PORT)
    print(f"\nList connections:")
    print(connections.list_connections())
    connections.disconnect('test')

async def get_agent(agent_name: str):
    """Return the New Agent.
    """
    llm = GroqLLM(
        model="llama3-groq-70b-8192-tool-use-preview",
        temperature=0.1,
        top_k=30,
        Top_p=0.6,
    )
    agent = Chatbot(
        name=agent_name,
        llm=llm
    )
    await agent.configure()
    return agent

if __name__ == "__main__":
    # Test Connections created by bot
    create_connection()
    loop = asyncio.get_event_loop()
    agent = loop.run_until_complete(
        get_agent('RFPBot')
    )
    print(
        'Agent: ', agent, agent.chatbot_id
    )
    create_connection()



import asyncio


def install_uvloop():
    """install uvloop and set as default loop for asyncio."""
    try:
        import uvloop  # noqa # pylint: disable=import-outside-toplevel
        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
        uvloop.install()
    except ImportError:
        pass



from typing import Optional, Union
from transformers import pipeline
import torch
from langchain.docstore.document import Document
from .youtube import YoutubeLoader


class VimeoLoader(YoutubeLoader):
    """
    Loader for Vimeo videos.
    """
    def load_video(self, url: str, video_title: str, transcript: Optional[Union[str, None]] = None) -> list:
        metadata = {
            "source": url,
            "url": url,
            "index": url,
            "filename": video_title,
            "question": '',
            "answer": '',
            'type': 'video_transcript',
            "source_type": self._source_type,
            "summary": '',
            "document_meta": {
                "language": self._language,
                "title": video_title,
                "topic_tags": ""
            }
        }
        if self.topics:
            metadata['document_meta']['topic_tags'] = self.topics
        if transcript is None:
            documents = []
            docs = []
            # first: download video
            try:
                file_path = self.download_video(url, self._video_path)
            except Exception:
                return []
            if not file_path:
                self.logger.warning(
                    f"Error downloading File for video: {self._video_path}"
                )
                return []
            transcript_path = file_path.with_suffix('.vtt')
            audio_path = file_path.with_suffix('.mp3')
            # second: extract audio
            self.extract_audio(file_path, audio_path)
            # get the Whisper parser
            transcript_whisper = self.get_whisper_transcript(audio_path)
            if transcript_whisper:
                transcript = transcript_whisper['text']
            else:
                transcript = ''
            # Summarize the transcript
            if transcript:
                summary = self.get_summary_from_text(transcript)
                # Create Two Documents, one is for transcript, second is VTT:
                metadata['summary'] = summary
                doc = Document(
                    page_content=transcript,
                    metadata=metadata
                )
                documents.append(doc)
            if transcript_whisper:
                # VTT version:
                transcript = self.transcript_to_vtt(transcript_whisper, transcript_path)
                doc = Document(
                    page_content=transcript,
                    metadata=metadata
                )
                documents.append(doc)
                # Saving every dialog chunk as a separate document
                dialogs = self.transcript_to_blocks(transcript_whisper)
                for chunk in dialogs:
                    _meta = {
                        "index": f"{video_title}:{chunk['id']}",
                        "document_meta": {
                            "start": f"{chunk['start_time']}",
                            "end": f"{chunk['end_time']}",
                            "id": f"{chunk['id']}",
                            "language": self._language,
                            "title": video_title,
                            "topic_tags": ""
                        }
                    }
                    _info = {**metadata, **_meta}
                    doc = Document(
                        page_content=chunk['text'],
                        metadata=_info
                    )
                    docs.append(doc)
            documents.extend(docs)
            return self.split_documents(documents)
        else:
            # using the transcript file
            with open(transcript, 'r') as f:
                transcript = f.read()
                summary = self.get_summary_from_text(transcript)
                transcript_whisper = None
            metadata['summary'] = f"{summary!s}"
            # Create Two Documents, one is for transcript, second is VTT:
            doc = Document(
                page_content=transcript,
                metadata=metadata
            )
            return self.split_documents([doc])



"""
Chatbot Manager.

Tool for instanciate, managing and interacting with Chatbot through APIs.
"""
from typing import Any, Dict, Type
from importlib import import_module
from aiohttp import web
from navconfig.logging import logging
from .chatbots import (
    AbstractChatbot,
    Chatbot
)
from .handlers.chat import ChatHandler # , BotHandler
from .handlers import ChatbotHandler
from .models import ChatbotModel
# Manual Load of Copilot Agent:
from parrot.chatbots.copilot import CopilotAgent

class ChatbotManager:
    """ChatbotManager.

    Manage chatbots and interact with them through via aiohttp App.

    """
    app: web.Application = None
    chatbots: Dict[str, AbstractChatbot] = {}

    def __init__(self) -> None:
        self.app = None
        self.chatbots = {}
        self.logger = logging.getLogger(name='Parrot.Manager')

    def get_chatbot_class(self, class_name: str) -> Type[AbstractChatbot]:
        """
        Dynamically import a chatbot class based on the class name from the relative module '.chatbots'.
        Args:
        class_name (str): The name of the chatbot class to be imported.
        Returns:
        Type[AbstractChatbot]: A chatbot class derived from AbstractChatbot.
        """
        module = import_module('.chatbots', __package__)
        try:
            return getattr(module, class_name)
        except AttributeError:
            raise ImportError(f"No class named '{class_name}' found in the module 'chatbots'.")

    async def load_bots(self, app: web.Application) -> None:
        """Load all chatbots from DB."""
        self.logger.info("Loading chatbots from DB...")
        db = app['database']
        async with await db.acquire() as conn:
            ChatbotModel.Meta.connection = conn
            bots = await ChatbotModel.all()
            for bot in bots:
                if bot.bot_type == 'chatbot':
                    self.logger.notice(
                        f"Loading chatbot '{bot.name}'..."
                    )
                    cls_name = bot.custom_class
                    if cls_name is None:
                        class_name = Chatbot
                    else:
                        class_name = self.get_chatbot_class(cls_name)
                    chatbot = class_name(
                        chatbot_id=bot.chatbot_id,
                        name=bot.name
                    )
                    try:
                        await chatbot.configure()
                    except Exception as e:
                        self.logger.error(
                            f"Failed to configure chatbot '{chatbot.name}': {e}"
                        )
                elif bot.bot_type == 'agent':
                    self.logger.notice(
                        f"Loading Agent '{bot.name}'..."
                    )
                    # TODO: extract the list of tools from Agent config
                    try:
                        tools = CopilotAgent.default_tools()
                        chatbot = CopilotAgent(
                            name=bot.name,
                            llm=bot.llm,
                            tools=tools
                        )
                    except Exception as e:
                        print('AQUI >>> ', e)
                        self.logger.error(
                            f"Failed to configure Agent '{bot.name}': {e}"
                        )
                self.add_chatbot(chatbot)
        self.logger.info(
            ":: Chatbots loaded successfully."
        )

    def create_chatbot(self, class_name: Any = None, name: str = None, **kwargs) -> AbstractChatbot:
        """Create a chatbot and add it to the manager."""
        if class_name is None:
            class_name = Chatbot
        chatbot = class_name(**kwargs)
        chatbot.name = name
        self.add_chatbot(chatbot)
        if 'llm' in kwargs:
            llm = kwargs['llm']
            llm_name = llm.pop('name')
            model = llm.pop('model')
            llm = chatbot.load_llm(
                llm_name, model=model, **llm
            )
            chatbot.llm = llm
        return chatbot

    def add_chatbot(self, chatbot: AbstractChatbot) -> None:
        """Add a chatbot to the manager."""
        self.chatbots[chatbot.name] = chatbot

    def get_chatbot(self, name: str) -> AbstractChatbot:
        """Get a chatbot by name."""
        return self.chatbots.get(name)

    def remove_chatbot(self, name: str) -> None:
        """Remove a chatbot by name."""
        del self.chatbots[name]

    def get_chatbots(self) -> Dict[str, AbstractChatbot]:
        """Get all chatbots."""
        return self.chatbots

    def get_app(self) -> web.Application:
        """Get the app."""
        if self.app is None:
            raise RuntimeError("App is not set.")
        return self.app

    def setup(self, app: web.Application) -> web.Application:
        if isinstance(app, web.Application):
            self.app = app  # register the app into the Extension
        else:
            self.app = app.get_app()  # Nav Application
        # register signals for startup and shutdown
        self.app.on_startup.append(self.on_startup)
        self.app.on_shutdown.append(self.on_shutdown)
        # Add Manager to main Application:
        self.app['chatbot_manager'] = self
        ## Configure Routes
        router = self.app.router
        router.add_view(
            '/api/v1/chat',
            ChatHandler
        )
        router.add_view(
            '/api/v1/chat/{chatbot_name}',
            ChatHandler
        )
        ChatbotHandler.configure(self.app, '/api/v1/bots')
        return self.app

    async def on_startup(self, app: web.Application) -> None:
        """On startup."""
        # configure all pre-configured chatbots:
        await self.load_bots(app)

    async def on_shutdown(self, app: web.Application) -> None:
        """On shutdown."""
        pass



from collections.abc import Callable
from pathlib import Path, PurePath
from typing import Any
from io import BytesIO
import re
import ftfy
import fitz
import pytesseract
from paddleocr import PaddleOCR
import torch
import cv2
from transformers import (
    # DonutProcessor,
    # VisionEncoderDecoderModel,
    # VisionEncoderDecoderConfig,
    # ViTImageProcessor,
    # AutoTokenizer,
    LayoutLMv3ForTokenClassification,
    LayoutLMv3Processor
)
from pdf4llm import to_markdown
from PIL import Image
from langchain.docstore.document import Document
from navconfig import config
from .basepdf import BasePDF


class PDFLoader(BasePDF):
    """
    Loader for PDF files.
    """
    def __init__(
        self,
        path: PurePath,
        tokenizer: Callable[..., Any] = None,
        text_splitter: Callable[..., Any] = None,
        source_type: str = 'pdf',
        language: str = "eng",
        **kwargs
    ):
        super().__init__(
            path=path,
            tokenizer=tokenizer,
            text_splitter=text_splitter,
            source_type=source_type,
            language=language,
            **kwargs
        )
        self.parse_images = kwargs.get('parse_images', False)
        self.page_as_images = kwargs.get('page_as_images', False)
        if self.page_as_images is True:
            # # Load the processor and model from Hugging Face
            # self.image_processor = DonutProcessor.from_pretrained(
            #     "naver-clova-ix/donut-base-finetuned-docvqa"
            # )
            # self.image_model = VisionEncoderDecoderModel.from_pretrained(
            #     "naver-clova-ix/donut-base-finetuned-docvqa",

            # )
            # Load the processor and model from Hugging Face
            self.image_processor = LayoutLMv3Processor.from_pretrained(
                "microsoft/layoutlmv3-base",
                apply_ocr=True
            )
            self.image_model = LayoutLMv3ForTokenClassification.from_pretrained(
                # "microsoft/layoutlmv3-base-finetuned-funsd"
                "HYPJUDY/layoutlmv3-base-finetuned-funsd"
            )
            # Set device to GPU if available
            self.image_device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            self.image_model.to(self.image_device)

        # Table Settings:
        self.table_settings = {
            #"vertical_strategy": "text",
            # "horizontal_strategy": "text",
            "intersection_x_tolerance": 3,
            "intersection_y_tolerance": 3
        }
        table_settings = kwargs.get('table_setttings', {})
        if table_settings:
            self.table_settings.update(table_settings)

    def explain_image(self, image_path):
        """Function to explain the image."""
        # with open(image_path, "rb") as image_file:
        #     image_content = image_file.read()

        # Open the image
        image = cv2.imread(image_path)
        task_prompt = "<s_docvqa><s_question>{user_input}</s_question><s_answer>"
        question = "Extract Questions about Happily Greet"
        prompt = task_prompt.replace("{user_input}", question)

        decoder_input_ids = self.image_processor.tokenizer(
            prompt,
            add_special_tokens=False,
            return_tensors="pt",
        ).input_ids

        pixel_values = self.image_processor(
            image,
            return_tensors="pt"
        ).pixel_values

        # Send inputs to the appropriate device
        pixel_values = pixel_values.to(self.image_device)
        decoder_input_ids = decoder_input_ids.to(self.image_device)

        outputs = self.image_model.generate(
            pixel_values,
            decoder_input_ids=decoder_input_ids,
            max_length=self.image_model.decoder.config.max_position_embeddings,
            pad_token_id=self.image_processor.tokenizer.pad_token_id,
            eos_token_id=self.image_processor.tokenizer.eos_token_id,
            bad_words_ids=[[self.image_processor.tokenizer.unk_token_id]],
            # use_cache=True
            return_dict_in_generate=True,
        )

        sequence = self.image_processor.batch_decode(outputs.sequences)[0]


        sequence = sequence.replace(
            self.image_processor.tokenizer.eos_token, ""
        ).replace(
            self.image_processor.tokenizer.pad_token, ""
        )
        # remove first task start token
        sequence = re.sub(r"<.*?>", "", sequence, count=1).strip()
        # Print the extracted sequence
        print("Extracted Text:", sequence)

        print(self.image_processor.token2json(sequence))

        # Format the output as Markdown (optional step)
        markdown_text = self.format_as_markdown(sequence)
        print("Markdown Format:\n", markdown_text)

        return None

    def convert_to_markdown(self, text):
        """
        Convert the cleaned text into a markdown format.
        You can enhance this function to detect tables, headings, etc.
        """
        # For example, we can identify sections or headers and format them in Markdown
        markdown_text = text
        # Detect headings and bold them
        markdown_text = re.sub(r"(^.*Scorecard.*$)", r"## \1", markdown_text)
        # Convert lines with ":" to a list item (rough approach)
        markdown_text = re.sub(r"(\w+):", r"- **\1**:", markdown_text)
        # Return the markdown formatted text
        return markdown_text

    def clean_tokenized_text(self, tokenized_text):
        """
        Clean the tokenized text by fixing encoding issues and formatting, preserving line breaks.
        """
        # Fix encoding issues using ftfy
        cleaned_text = ftfy.fix_text(tokenized_text)

        # Remove <s> and </s> tags (special tokens)
        cleaned_text = cleaned_text.replace("<s>", "").replace("</s>", "")

        # Replace special characters like 'Ġ' and fix multiple spaces, preserving new lines
        cleaned_text = cleaned_text.replace("Ġ", " ")

        # Avoid collapsing line breaks, but still normalize multiple spaces
        # Replace multiple spaces with a single space, but preserve line breaks
        cleaned_text = re.sub(r" +", " ", cleaned_text)

        return cleaned_text.strip()

    def extract_page_text(self, image_path) -> str:
        # Open the image
        image = Image.open(image_path).convert("RGB")

        # Processor handles the OCR internally, no need for words or boxes
        encoding = self.image_processor(image, return_tensors="pt", truncation=True)
        encoding = {k: v.to(self.image_device) for k, v in encoding.items()}

        # Forward pass
        outputs = self.image_model(**encoding)
        logits = outputs.logits

        # Get predictions
        predictions = logits.argmax(-1).squeeze().tolist()
        labels = [self.image_model.config.id2label[pred] for pred in predictions]

        # Get the words and boxes from the processor's OCR step
        words = self.image_processor.tokenizer.convert_ids_to_tokens(
            encoding['input_ids'].squeeze().tolist()
        )
        boxes = encoding['bbox'].squeeze().tolist()

        # Combine words and labels, preserving line breaks based on vertical box position
        extracted_text = ""
        last_box = None
        for word, label, box in zip(words, labels, boxes):
            if label != 'O':
                # Check if the current word is on a new line based on the vertical position of the box
                if last_box and abs(box[1] - last_box[1]) > 10:  # A threshold for line breaks
                    extracted_text += "\n"  # Add a line break

                extracted_text += f"{word} "
                last_box = box
        cleaned_text = self.clean_tokenized_text(extracted_text)
        markdown_text = self.convert_to_markdown(cleaned_text)
        return markdown_text

    def _load_pdf(self, path: Path) -> list:
        """
        Load a PDF file using the Fitz library.

        Args:
            path (Path): The path to the PDF file.

        Returns:
            list: A list of Langchain Documents.
        """
        if self._check_path(path):
            self.logger.info(f"Loading PDF file: {path}")
            pdf = fitz.open(str(path))  # Open the PDF file
            docs = []
            try:
                md_text = to_markdown(pdf) # get markdown for all pages
                _meta = {
                    "url": f'{path}',
                    "source": f"{path.name}",
                    "filename": path.name,
                    "type": 'pdf',
                    "question": '',
                    "answer": '',
                    "source_type": self._source_type,
                    "data": {},
                    "summary": '',
                    "document_meta": {
                        "title": pdf.metadata.get("title", ""),
                        "creationDate": pdf.metadata.get("creationDate", ""),
                        "author": pdf.metadata.get("author", ""),
                    }
                }
                docs.append(
                    Document(
                        page_content=md_text,
                        metadata=_meta
                    )
                )
            except Exception:
                pass
            for page_number in range(pdf.page_count):
                page = pdf[page_number]
                text = page.get_text()
                # first: text
                if text:
                    page_num = page_number + 1
                    try:
                        summary = self.get_summary_from_text(text)
                    except Exception:
                        summary = ''
                    metadata = {
                        "url": '',
                        "source": f"{path.name} Page.#{page_num}",
                        "filename": path.name,
                        "index": f"{page_num}",
                        "type": 'pdf',
                        "question": '',
                        "answer": '',
                        "source_type": self._source_type,
                        "data": {},
                        "summary": summary,
                        "document_meta": {
                            "title": pdf.metadata.get("title", ""),
                            "creationDate": pdf.metadata.get("creationDate", ""),
                            "author": pdf.metadata.get("author", ""),
                        }
                    }
                    docs.append(
                        Document(
                            page_content=text,
                            metadata=metadata
                        )
                    )
                # Extract images and use OCR to get text from each image
                # second: images
                file_name = path.stem.replace(' ', '_').replace('.', '').lower()
                if self.parse_images is True:
                    # extract any images in page:
                    image_list = page.get_images(full=True)
                    for img_index, img in enumerate(image_list):
                        xref = img[0]
                        base_image = pdf.extract_image(xref)
                        image = Image.open(BytesIO(base_image["image"]))
                        url = ''
                        if self.save_images is True:
                            img_name = f'image_{file_name}_{page_num}_{img_index}.png'
                            img_path = self._imgdir.joinpath(img_name)
                            self.logger.notice(
                                f"Saving Image Page on {img_path}"
                            )
                            try:
                                image.save(
                                    img_path,
                                    format="png",
                                    optimize=True
                                )
                                url = f'/static/images/{img_name}'
                            except OSError:
                                pass
                        # Use Tesseract to extract text from image
                        image_text = pytesseract.image_to_string(
                            image,
                            lang=self._lang
                        )
                        # TODO: add the summary (explanation)
                        # Create a document for each image
                        image_meta = {
                            "url": url,
                            "source": f"{path.name} Page.#{page_num}",
                            "filename": path.name,
                            "index": f"{path.name}:{page_num}",
                            "question": '',
                            "answer": '',
                            "type": 'image',
                            "data": {},
                            "summary": '',
                            "document_meta": {
                                "image_index": img_index,
                                "image_name": img_name,
                                "description": f"Extracted from {page_number}."
                            },
                            "source_type": self._source_type
                        }
                        docs.append(
                            Document(page_content=image_text, metadata=image_meta)
                        )
                # third: tables
                # Look for tables on this page and display the table count
                try:
                    tabs = page.find_tables()
                    for tab_idx, tab in enumerate(tabs):
                        # iterating over all tables in page:
                        df = tab.to_pandas()  # convert to pandas DataFrame
                        # converting to markdown, but after pre-processing pandas
                        df = df.dropna(axis=1, how='all')
                        df = df.dropna(how='all', axis=0)  # Drop empty rows
                        table_meta = {
                            "url": '',
                            "source": f"{path.name} Page.#{page_num} Table.#{tab_idx}",
                            "filename": path.name,
                            "index": f"{path.name}:{page_num}",
                            "question": '',
                            "answer": '',
                            "type": 'table',
                            "data": {},
                            "summary": '',
                            "document_meta": {
                                "table_index": tab_idx,
                                "table_shape": df.shape,
                                "table_columns": df.columns.tolist(),
                                "description": f"Extracted from {page_number}."
                            },
                            "source_type": self._source_type
                        }
                        txt = df.to_markdown()
                        if txt:
                            docs.append(
                                Document(page_content=txt, metadata=table_meta)
                            )
                except Exception as exc:
                    print(exc)
                # fourth: page as image
                if self.page_as_images is True:
                    # Convert the page to a Pixmap (which is an image)
                    mat = fitz.Matrix(2, 2)
                    pix = page.get_pixmap(dpi=300, matrix=mat) # Increase DPI for better resolution
                    img_name = f'{file_name}_page_{page_num}.png'
                    img_path = self._imgdir.joinpath(img_name)
                    if img_path.exists():
                        img_path.unlink(missing_ok=True)
                    self.logger.notice(
                        f"Saving Page {page_number} as Image on {img_path}"
                    )
                    pix.save(
                        img_path
                    )
                    # TODO passing the image to a AI visual to get explanation
                    # Get the extracted text from the image
                    text = self.extract_page_text(img_path)
                    url = f'/static/images/{img_name}'
                    image_meta = {
                        "url": url,
                        "source": f"{path.name} Page.#{page_num}",
                        "filename": path.name,
                        "index": f"{path.name}:{page_num}",
                        "question": '',
                        "answer": '',
                        "type": 'page',
                        "data": {},
                        "summary": '',
                        "document_meta": {
                            "image_name": img_name,
                            "page_number": f"{page_number}"
                        },
                        "source_type": self._source_type
                    }
                    docs.append(
                        Document(page_content=text, metadata=image_meta)
                    )
            pdf.close()
            return docs
        else:
            return []

    def get_ocr(self, img_path) -> list:
        # Initialize PaddleOCR with table recognition
        self.ocr_model = PaddleOCR(
            lang='en',
            det_model_dir=None,
            rec_model_dir=None,
            rec_char_dict_path=None,
            table=True,
            # use_angle_cls=True,
            # use_gpu=True
        )
        result = self.ocr_model.ocr(img_path, cls=True)

        # extract tables:
        # The result contains the table structure and content
        tables = []
        for line in result:
            if 'html' in line[1]:
                html_table = line[1]['html']
                tables.append(html_table)

        print('TABLES > ', tables)



from .parsers import TOMLParser


async def parse_toml_config(config_dir: str) -> dict:
    try:
        parser = TOMLParser()
        return await parser.parse(config_dir)
    except Exception as exc:
        raise ValueError(
            f"Error Parsing TOML Config on {config_dir}: {exc}"
        )



from typing import Any
from collections.abc import Callable
from pathlib import Path, PurePath
from langchain_community.document_loaders import TextLoader
from langchain.docstore.document import Document
from .abstract import AbstractLoader


class TXTLoader(AbstractLoader):
    """
    Loader for PDF files.
    """
    _extension = ['.txt']

    def __init__(
        self,
        path: PurePath,
        tokenizer: Callable[..., Any] = None,
        text_splitter: Callable[..., Any] = None,
        source_type: str = 'text',
        **kwargs
    ):
        super().__init__(tokenizer, text_splitter, source_type=source_type, **kwargs)
        self.path = path
        if isinstance(path, str):
            self.path = Path(path).resolve()

    def _load_document(self, path: Path) -> list:
        """
        Load a TXT file.

        Args:
            path (Path): The path to the TXT file.

        Returns:
            list: A list of Langchain Documents.
        """
        if self._check_path(path):
            self.logger.info(f"Loading TXT file: {path}")
            with open(path, 'r') as file:
                text = file.read()
            return [
                Document(
                    page_content=text,
                    metadata={
                        "url": '',
                        "index": str(path.name),
                        "source": str(path),
                        "filename": str(path.name),
                        "summary": '',
                        "question": '',
                        "answer": '',
                        'type': 'text',
                        "source_type": self._source_type,
                        "document_meta": {}
                    }
                )
            ]
        return []

    def load(self) -> list:
        """
        Load data from a TXT file.

        Args:
            source (str): The path to the TXT file.

        Returns:
            list: A list of Langchain Documents.
        """
        if not self.path.exists():
            raise FileNotFoundError(
                f"File/directory not found: {self.path}"
            )
        if self.path.is_dir():
            documents = []
            # iterate over the files in the directory
            for ext in self._extension:
                for item in self.path.glob(f'*{ext}'):
                    documents.extend(self._load_document(item))
        elif self.path.is_file():
            documents = self._load_document(self.path)
        else:
            raise ValueError(
                f"TXT Loader: Invalid path: {self.path}"
            )
        return self.split_documents(documents)

    def parse(self, source):
        raise NotImplementedError(
            "Parser method is not implemented for TXTLoader."
        )



from collections.abc import Callable
from pathlib import PurePath
from langchain_community.document_loaders import JSONLoader as JSLoader
from .abstract import AbstractLoader


class JSONLoader(AbstractLoader):
    """
    Loader for JSON files.
    """
    _extension = ['.json']
    extract_metadata: Callable = None

    def extract_metadata(self, record: dict, metadata: dict) -> dict:
        meta = {
            "source_type": self._source_type,
            "priority": self._priority,
        }
        return meta

    def load(self, path: PurePath) -> list:
        """
        Load data from a JSON file.

        Args:
            source (str): The path to the JSON file.

        Returns:
            list: A list of Langchain Documents.
        """
        if self._check_path(path):
            self.logger.info(f"Loading JSON file: {path}")
            # Create metadata for each chunk
            meta = {
                "filename": str(path),
            }
            args = {
                "metadata_func": self.extract_metadata,
            }
            loader = JSLoader(
                file_path=path,
                jq_schema=".",
                text_content=False,
                **args
            )
            documents = loader.load()
            for doc in documents:
                doc.metadata.update(meta)
            # Split the documents into chunks
            return self.split_documents(documents)
        else:
            return []



#!/usr/bin/env python3
import asyncio
import time
from functools import partial
from aiohttp import web
from navigator import Application
from app import Main
app = Application(Main, enable_jinja2=True)


async def send_email(email, message):
    print(' :: Waiting for 10 seconds to finish task :: ')
    await asyncio.sleep(10)  # Simulate email sending
    print(f"Email sent to {email} with message: {message}")

def blocking_code(request):
    time.sleep(10)
    print(":: Blocking code executed ::")

async def handle_post(request):
    data = await request.json()
    tasker = request.app['tasker']
    await tasker.put(send_email, data['email'], data['message'])
    fn = partial(blocking_code, request)
    await tasker.put(fn)
    return web.json_response({'status': 'Task enqueued'})


app.router.add_post('/send_email', handle_post)


if __name__ == '__main__':
    try:
        app.run()
    except KeyboardInterrupt:
        print(
            "Closing Parrot Chatbot Service ..."
        )



from typing import Any
from .abstract import AbstractStore
from ..exceptions import ConfigError  # pylint: disable=E0611
try:
    from .qdrant import QdrantStore
    QDRANT_ENABLED = True
except (ModuleNotFoundError, ImportError):
    QDRANT_ENABLED = False

try:
    from .milvus import MilvusStore
    MILVUS_ENABLED = True
except (ModuleNotFoundError, ImportError):
    MILVUS_ENABLED = False


def get_vectordb(vector_db: str, embeddings: Any, **kwargs) -> AbstractStore:
    if vector_db in ('QdrantStore', 'qdrant'):
        if QDRANT_ENABLED is True:
            ## TODO: support pluggable vector store
            return QdrantStore(  # pylint: disable=E0110
                embeddings=embeddings,
                **kwargs
            )
        else:
            raise ConfigError(
                (
                    "Qdrant is enabled but not installed, "
                    "Hint: Please install with pip install -e .[qdrant]"
                )
            )
    if vector_db in ('milvus', 'MilvusStore'):
        if MILVUS_ENABLED is True:
            return MilvusStore(
                embeddings=embeddings,
                **kwargs
            )
        else:
            raise ConfigError(
                (
                    "Milvus is enabled but not installed, "
                    "Hint: Please install with pip install -e .[milvus]"
                )
            )
    else:
        raise ValueError(
            f"Vector Database {vector_db} not supported"
        )



"""Loaders are classes that are responsible for loading data from a source
and returning it as a Langchain Document.
"""
from __future__ import annotations
from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import List, Union, Optional, Any
from pathlib import Path, PurePath
import torch
from langchain.docstore.document import Document
from langchain.chains.summarize import load_summarize_chain
from langchain.text_splitter import (
    RecursiveCharacterTextSplitter,
    TokenTextSplitter
)
from langchain_core.prompts import PromptTemplate
from langchain_core.document_loaders.blob_loaders import Blob
from langchain_community.llms.huggingface_pipeline import HuggingFacePipeline

from navconfig.logging import logging
from navigator.libs.json import JSONContent  # pylint: disable=E0611
from transformers import (
    AutoTokenizer,
    AutoModelForSeq2SeqLM,
    AutoConfig,
    AutoModel,
    pipeline
)
from ..conf import EMBEDDING_DEVICE, EMBEDDING_DEFAULT_MODEL


logging.getLogger(name='httpx').setLevel(logging.WARNING)
logging.getLogger(name='httpcore').setLevel(logging.WARNING)
logging.getLogger(name='pdfminer').setLevel(logging.WARNING)
logging.getLogger(name='langchain_community').setLevel(logging.WARNING)
logging.getLogger(name='numba').setLevel(logging.WARNING)
logging.getLogger(name='PIL').setLevel(level=logging.WARNING)


def as_string(self) -> str:
    """Read data as a string."""
    if self.data is None and self.path:
        with open(str(self.path), "r", encoding=self.encoding) as f:
            try:
                return f.read()
            except UnicodeDecodeError:
                try:
                    with open(str(self.path), "r", encoding="latin-1") as f:
                        return f.read()
                except UnicodeDecodeError:
                    with open(str(self.path), "rb") as f:
                        return f.read().decode("utf-8", "replace")
    elif isinstance(self.data, bytes):
        return self.data.decode(self.encoding)
    elif isinstance(self.data, str):
        return self.data
    else:
        raise ValueError(f"Unable to get string for blob {self}")

# Monkey patch the Blob class's as_string method
Blob.as_string = as_string


class AbstractLoader(ABC):
    """
    Abstract class for Document loaders.
    """
    _extension: List[str] = ['.txt']
    encoding: str = 'utf-8'
    skip_directories: List[str] = []
    _chunk_size: int = 768

    def __init__(
        self,
        tokenizer: Union[str, Callable] = None,
        text_splitter: Union[str, Callable] = None,
        source_type: str = 'file',
        **kwargs
    ):
        self.tokenizer = tokenizer
        self._summary_model = None
        self.text_splitter = text_splitter
        self._device = self._get_device()
        self._chunk_size = kwargs.get('chunk_size', 768)
        self._no_summarization = bool(
            kwargs.get('no_summarization', False)
        )
        self.summarization_model = kwargs.get(
            'summarization_model',
            "facebook/bart-large-cnn"
        )
        self._no_summarization = bool(
            kwargs.get('no_summarization', False)
        )
        self._source_type = source_type
        self.logger = logging.getLogger(
            f"Loader.{self.__class__.__name__}"
        )
        if 'extension' in kwargs:
            self._extension = kwargs['extension']
        self.encoding = kwargs.get('encoding', 'utf-8')
        self.skip_directories: List[str] = kwargs.get('skip_directories', [])
        # LLM (if required)
        self._llm = kwargs.get('llm', None)
        if not self.tokenizer:
            self.tokenizer = self.default_tokenizer()
        elif isinstance(self.tokenizer, str):
            self.tokenizer = self.get_tokenizer(
                model_name=self.tokenizer
            )
        if not text_splitter:
            self.text_splitter = self.default_splitter(
                model=self.tokenizer
            )
        # JSON encoder:
        self._encoder = JSONContent()


    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.post_load()

    def default_tokenizer(self):
        return self.get_tokenizer(
            EMBEDDING_DEFAULT_MODEL,
            chunk_size=768
        )

    def get_tokenizer(self, model_name: str, chunk_size: int = 768):
        return AutoTokenizer.from_pretrained(
            model_name,
            chunk_size=chunk_size
        )

    def _get_device(self, cuda_number: int = 0):
        if torch.cuda.is_available():
            # Use CUDA GPU if available
            device = torch.device(f'cuda:{cuda_number}')
        elif torch.backends.mps.is_available():
            # Use CUDA Multi-Processing Service if available
            device = torch.device("mps")
        elif EMBEDDING_DEVICE == 'cuda':
            device = torch.device(f'cuda:{cuda_number}')
        else:
            device = torch.device(EMBEDDING_DEVICE)
        return device

    def get_model(self, model_name: str):
        self._model_config = AutoConfig.from_pretrained(
            model_name, trust_remote_code=True
        )
        return AutoModel.from_pretrained(
            model_name,
            trust_remote_code=True,
            config=self._model_config,
            unpad_inputs=True,
            use_memory_efficient_attention=True,
        ).to(self._device)

    def get_summarization_model(self, model_name: str = 'facebook/bart-large-cnn'):
        if self._no_summarization is True:
            return None
        if not self._summary_model:
            summarize_model = AutoModelForSeq2SeqLM.from_pretrained(
                model_name,
                device_map="auto",
                torch_dtype=torch.bfloat16,
                trust_remote_code=True
            )
            summarize_tokenizer = AutoTokenizer.from_pretrained(
                model_name,
                padding_side="left"
            )
            pipe_summary = pipeline(
                "summarization",
                model=summarize_model,
                tokenizer=summarize_tokenizer,
                batch_size=True,
                max_new_tokens=500,
                min_new_tokens=300,
                use_fast=True
            )
            self._summary_model = HuggingFacePipeline(
                model_id=model_name,
                pipeline=pipe_summary,
                verbose=True
            )
        return self._summary_model

    def get_text_splitter(self, model, chunk_size: int = 2000, overlap: int = 100):
        return RecursiveCharacterTextSplitter.from_huggingface_tokenizer(
            model,
            chunk_size=chunk_size,
            chunk_overlap=overlap,
            add_start_index=True,  # If `True`, includes chunk's start index in metadata
            strip_whitespace=True,  # strips whitespace from the start and end
            separators=["\n\n", "\n", "\r\n", "\r", "\f", "\v", "\x0b", "\x0c"],
        )

    def default_splitter(self, model: Callable):
        """Get the text splitter."""
        return self.get_text_splitter(
            model,
            chunk_size=2000,
            overlap=100
        )

    def get_summary_from_text(self, text: str) -> str:
        """
        Get a summary of a text.
        """
        if not text:
            # NO data to be summarized
            return ''
        try:
            splitter = TokenTextSplitter(
                chunk_size=5000,
                chunk_overlap=100,
            )
            prompt_template = """Write a summary of the following, please also identify the main theme:
            {text}
            SUMMARY:"""
            prompt = PromptTemplate.from_template(prompt_template)
            refine_template = (
                "Your job is to produce a final summary\n"
                "We have provided an existing summary up to a certain point: {existing_answer}\n"
                "We have the opportunity to refine the existing summary"
                "(only if needed) with some more context below.\n"
                "------------\n"
                "{text}\n"
                "------------\n"
                "Given the new context, refine the original summary adding more explanation."
                "If the context isn't useful, return the original summary."
            )
            refine_prompt = PromptTemplate.from_template(refine_template)
            if self._llm:
                llm = self._llm
            else:
                llm = self.get_summarization_model(
                    self.summarization_model
                )
                if not llm:
                    return ''
            summarize_chain = load_summarize_chain(
                llm=llm,
                chain_type="refine",
                question_prompt=prompt,
                refine_prompt=refine_prompt,
                return_intermediate_steps=True,
                input_key="input_documents",
                output_key="output_text",
            )
            chunks = splitter.split_text(text)
            documents = [Document(page_content=chunk) for chunk in chunks]
            summary = summarize_chain.invoke(
                {"input_documents": documents}, return_only_outputs=True
            )
            return summary['output_text']
        except Exception as e:
            print('ERROR in get_summary_from_text:', e)
            return ""

    def split_documents(self, documents: List[Document], max_tokens: int = None) -> List[Document]:
        """Split the documents into chunks."""
        if not max_tokens:
            max_tokens = self._chunk_size
        split_documents = []
        for doc in documents:
            metadata = doc.metadata.copy()
            chunks = self.text_splitter.split_text(doc.page_content)
            for chunk in chunks:
                split_documents.append(
                    Document(page_content=chunk, metadata=metadata)
                )
        return split_documents

    def split_by_tokens(self, documents: List[Document], max_tokens: int = 768) -> List[Document]:
        """Split the documents into chunks."""
        split_documents = []
        current_chunk = []
        for doc in documents:
            metadata = doc.metadata.copy()
            tokens = self.tokenizer.tokenize(doc.page_content)
            with torch.no_grad():
                current_chunk = []
                for token in tokens:
                    current_chunk.append(token)
                    if len(current_chunk) >= max_tokens:
                        chunk_text = self.tokenizer.convert_tokens_to_string(current_chunk)
                        # Create a new Document for this chunk, preserving metadata
                        split_doc = Document(
                            page_content=chunk_text,
                            metadata=metadata
                        )
                        split_documents.append(split_doc)
                        current_chunk = []  # Reset for the next chunk
            # Handle the last chunk if it didn't reach the max_tokens limit
            if current_chunk:
                chunk_text = self.tokenizer.convert_tokens_to_string(current_chunk)
                split_documents.append(
                    Document(page_content=chunk_text, metadata=metadata)
                )
            del tokens, current_chunk
            torch.cuda.empty_cache()
        return split_documents

    def post_load(self):
        self.tokenizer = None # Reset the tokenizer
        self.text_splitter = None # Reset the text splitter
        torch.cuda.synchronize()  # Wait for all kernels to finish
        torch.cuda.empty_cache()  # Clear unused memory

    def read_bytes(self, path: Union[str, PurePath]) -> bytes:
        """Read the bytes from a file.

        Args:
            path (Union[str, PurePath]): The path to the file.

        Returns:
            bytes: The bytes of the file.
        """
        if isinstance(path, str):
            path = PurePath(path)
        with open(str(path), 'rb') as f:
            return f.read()

    def open_bytes(self, path: Union[str, PurePath]) -> Any:
        """Open the bytes from a file.

        Args:
            path (Union[str, PurePath]): The path to the file.

        Returns:
            Any: The bytes of the file.
        """
        if isinstance(path, str):
            path = PurePath(path)
        return open(str(path), 'rb')

    def read_string(self, path: Union[str, PurePath]) -> str:
        """Read the string from a file.

        Args:
            path (Union[str, PurePath]): The path to the file.

        Returns:
            str: The string of the file.
        """
        if isinstance(path, str):
            path = PurePath(path)
        with open(str(path), 'r', encoding=self.encoding) as f:
            return f.read()

    def _check_path(
        self,
        path: PurePath,
        suffix: Optional[List[str]] = None
    ) -> bool:
        """Check if the file path exists.
        Args:
            path (PurePath): The path to the file.
        Returns:
            bool: True if the file exists, False otherwise.
        """
        if isinstance(path, str):
            path = Path(path).resolve()
        if not suffix:
            suffix = self._extension
        return path.exists() and path.is_file() and path.suffix in suffix


    @abstractmethod
    def load(self, path: Union[str, PurePath]) -> List[Document]:
        """Load data from a source and return it as a Langchain Document.

        Args:
            path (str): The source of the data.

        Returns:
            List[Document]: A list of Langchain Documents.
        """
        pass

    @abstractmethod
    def parse(self, source: Any) -> List[Document]:
        """Parse data from a source and return it as a Langchain Document.

        Args:
            source (Any): The source of the data.

        Returns:
            List[Document]: A list of Langchain Documents.
        """
        pass

    @classmethod
    def from_path(
        cls,
        path: Union[str, PurePath],
        text_splitter: Callable,
        source_type: str = 'file',
        **kwargs
    ) -> List[Document]:
        """Load Multiple documents from a Path

        Args:
            path (Union[str, PurePath]): The path to the file.

        Returns:
            -> List[Document]: A list of Langchain Documents.
        """
        if isinstance(path, str):
            path = PurePath(path)
        if path.is_dir():
            documents = []
            obj = cls(
                tokenizer=kwargs.pop('tokenizer', None),
                text_splitter=text_splitter,
                source_type=source_type,
            )
            for ext in cls._extension:
                for item in path.glob(f'*{ext}'):
                    if set(item.parts).isdisjoint(cls.skip_directories):
                        documents += obj.load(path=item, **kwargs)
                        # documents += cls.load(cls, path=item, **kwargs)
            return documents

    @classmethod
    def from_url(
        cls,
        urls: List[str],
        text_splitter: Callable,
        source_type: str = 'content',
        **kwargs
    ) -> List[Document]:
        """Load Multiple documents from a URL

        Args:
            urls (List[str]): The list of URLs.

        Returns:
            -> List[Document]: A list of Langchain Documents.
        """
        documents = []
        cls.tokenizer=kwargs.pop('tokenizer', None),
        cls.text_splitter = text_splitter
        cls._source_type = source_type
        cls.summarization_model = kwargs.pop(
            'summarization_model',
            "facebook/bart-large-cnn"
        )
        for url in urls:
            documents += cls.load(url, **kwargs)
        return documents



from .abstract import AbstractChatbot
from .basic import Chatbot
from .asktroc import AskTROC
from .hragents import HRAgent
from .bose import BoseBot
from .odoo import OddieBot
from .cody import Cody



from typing import Optional, List
from datetime import datetime, timedelta
from uuid import UUID
from datamodel.types import Text
from asyncdb.models import Column, Model
from navigator_auth.conf import AUTH_DB_SCHEMA, AUTH_USERS_VIEW


class User(Model):
    """Basic User notation."""

    user_id: int = Column(
        required=False,
        primary_key=True,
        db_default="auto"
    )
    first_name: str
    last_name: str
    display_name: str
    email: str = Column(required=False, max=254)
    alt_email: str = Column(required=False, max=254)
    password: str = Column(required=False, max=128)
    last_login: datetime = Column(required=False)
    username: str = Column(required=False)
    is_superuser: bool = Column(required=True, default=False)
    is_active: bool = Column(required=True, default=True)
    is_new: bool = Column(required=True, default=True)
    is_staff: bool = Column(required=False, default=True)
    title: str = Column(equired=False, max=90)
    avatar: str = Column(max=512)
    associate_id: str = Column(required=False)
    associate_oid: str = Column(required=False)
    department_code: str = Column(required=False)
    job_code: str = Column(required=False)
    position_id: str = Column(required=False)
    group_id: list = Column(required=False)
    groups: list = Column(required=False)
    program_id: list = Column(required=False)
    programs: list = Column(required=False)
    start_date: datetime = Column(required=False)
    birthday: str = Column(required=False)
    worker_type: str = Column(required=False)
    created_at: datetime = Column(required=False)

    def birth_date(self):
        if self.birthday:
            _, month, day = self.birthday.split('-')
            # Get the current year
            current_year = datetime.now().year
            # Create a new date string with the current year
            new_date_str = f"{current_year}-{month}-{day}"
            # Convert the new date string to a datetime object
            return datetime.strptime(new_date_str, "%Y-%m-%d").date()
        return None

    def employment_duration(self):
        if not self.start_date:
            return None, None, None
        # Get today's date
        today = datetime.now().date()
        # employment:
        employment = self.start_date

        # Calculate the difference in years, months, days
        years = today.year - employment.year
        months = today.month - employment.month
        days = today.day - employment.day

        # Adjust for cases where the current month is before the start month
        if months < 0:
            years -= 1
            months += 12

        # Adjust for cases where the current day
        # is before the start day in the month
        if days < 0:
            # Subtract one month and calculate days based on the previous month
            months -= 1
            if months < 0:
                years -= 1
                months += 12
            # Calculate the last day of the previous month
            last_day_of_prev_month = (
                today.replace(day=1) - timedelta(days=1)
            ).day
            days += last_day_of_prev_month

        # Adjust months and years again if necessary
        if months < 0:
            years -= 1
            months += 12

        return years, months, days

    class Meta:
        driver = "pg"
        name = AUTH_USERS_VIEW
        schema = AUTH_DB_SCHEMA
        description = 'View Model for getting Users.'
        strict = True
        frozen = False


class UserIdentity(Model):

    identity_id: UUID = Column(
        required=False,
        primary_key=True,
        db_default="auto",
        repr=False
    )
    display_name: str = Column(required=False)
    title: str = Column(required=False)
    nickname: str = Column(required=False)
    email: str = Column(required=False)
    phone: str = Column(required=False)
    short_bio: Text = Column(required=False)
    avatar: str = Column(required=False)
    user_id: User = Column(required=False, repr=False)
    auth_provider: str = Column(required=False)
    auth_data: Optional[dict] = Column(required=False, repr=False)
    attributes: Optional[dict] = Column(required=False, repr=False)
    created_at: datetime = Column(
        required=False,
        default=datetime.now(),
        repr=False
    )

    class Meta:
        driver = "pg"
        name = "user_identities"
        description = 'Manage User Identities.'
        schema = AUTH_DB_SCHEMA
        strict = True

class ADUser(Model):
    """Active Directory Users."""
    people_id: UUID = Column(
        required=False,
        primary_key=True,
        db_default="auto",
        repr=False
    )
    username: str = Column(required=False)
    display_name: str = Column(required=False)
    given_name: str = Column(required=False)
    last_name: str = Column(required=False)
    phones: Optional[list] = Column(required=False)
    mobile: str = Column(required=False)
    job_title: str = Column(required=False)
    email: str = Column(required=False)
    office_location: str = Column(required=False)
    preferred_language: str = Column(required=False)
    associate_id: str = Column(required=False)
    associate_oid: str = Column(required=False)
    job_code_title: str = Column(required=False)
    position_id: str = Column(required=False)
    zammad_created: bool = Column(required=False, default=False)
    created_at: datetime = Column(
        required=False,
        default=datetime.now(),
        repr=False
    )

    class Meta:
        name = 'people'
        schema = 'troc'
        strict = True


class ADPeople(Model):
    """Active Directory Users."""
    people_id: UUID = Column(
        required=False,
        primary_key=True,
        db_default="auto",
        repr=False
    )
    user_id: int = Column(required=True)
    userid: UUID = Column(required=False)
    username: str = Column(required=False)
    display_name: str = Column(required=False)
    given_name: str = Column(required=False)
    last_name: str = Column(required=False)
    phones: Optional[list] = Column(required=False)
    mobile: str = Column(required=False)
    job_title: str = Column(required=False)
    email: str = Column(required=False)
    alt_email: str = Column(required=False)
    office_location: str = Column(required=False)
    preferred_language: str = Column(required=False)
    associate_id: str = Column(required=False)
    associate_oid: str = Column(required=False)
    job_code_title: str = Column(required=False)
    position_id: str = Column(required=False)
    created_at: datetime = Column(
        required=False,
        default=datetime.now(),
        repr=False
    )

    class Meta:
        name = 'vw_people'
        schema = 'troc'
        strict = True



from .abstract import AbstractChatbot


class Chatbot(AbstractChatbot):
    """Represents an Chatbot in Navigator.

        Each Chatbot has a name, a role, a goal, a backstory,
        and an optional language model (llm).
    """



from typing import Any
from collections.abc import Callable
from pathlib import PurePath
from langchain.docstore.document import Document
from .basevideo import BaseVideoLoader


class AudioLoader(BaseVideoLoader):
    """
    Generating transcripts from local Audio.
    """
    _extension = ['.mp3', '.webm']

    def __init__(
        self,
        path: PurePath,
        tokenizer: Callable[..., Any] = None,
        text_splitter: Callable[..., Any] = None,
        source_type: str = 'documentation',
        encoding: str = 'utf-8',
        origin: str = '',
        **kwargs
    ):
        super().__init__(tokenizer, text_splitter, source_type=source_type, **kwargs)
        self.path = path

    def load_audio(self, path: PurePath) -> list:
        metadata = {
            "source": f"{path}",
            "url": f"{path.name}",
            "index": path.stem,
            "filename": f"{path}",
            "question": '',
            "answer": '',
            'type': 'audio_transcript',
            "source_type": self._source_type,
            "summary": '',
            "document_meta": {
                "language": self._language,
                "topic_tags": ""
            }
        }
        documents = []
        transcript_path = path.with_suffix('.vtt')
        # get the Whisper parser
        transcript_whisper = self.get_whisper_transcript(path)
        if transcript_whisper:
            transcript = transcript_whisper['text']
        else:
            transcript = ''
        # Summarize the transcript
        if transcript:
            summary = self.get_summary_from_text(transcript)
            # Create Two Documents, one is for transcript, second is VTT:
            metadata['summary'] = summary
            doc = Document(
                page_content=transcript,
                metadata=metadata
            )
            documents.append(doc)
        if transcript_whisper:
            # VTT version:
            transcript = self.transcript_to_vtt(transcript_whisper, transcript_path)
            doc = Document(
                page_content=transcript,
                metadata=metadata
            )
            documents.append(doc)
            # Saving every dialog chunk as a separate document
            dialogs = self.transcript_to_blocks(transcript_whisper)
            docs = []
            for chunk in dialogs:
                _meta = {
                    "index": f"{path.stem}:{chunk['id']}",
                    "document_meta": {
                        "start": f"{chunk['start_time']}",
                        "end": f"{chunk['end_time']}",
                        "id": f"{chunk['id']}",
                        "language": self._language,
                        "title": f"{path.stem}",
                        "topic_tags": ""
                    }
                }
                _info = {**metadata, **_meta}
                doc = Document(
                    page_content=chunk['text'],
                    metadata=_info
                )
                docs.append(doc)
            documents.extend(docs)
        return documents

    def load(self) -> list:
        documents = []
        if self.path.is_file():
            docs = self.load_audio(self.path)
            documents.extend(docs)
        if self.path.is_dir():
            # iterate over the files in the directory
            for ext in self._extension:
                for item in self.path.glob(f'*{ext}'):
                    if set(item.parts).isdisjoint(self.skip_directories):
                        documents.extend(
                            self.load_audio(item)
                        )
        return self.split_documents(documents)



from typing import Any
from collections.abc import Callable
from pathlib import Path, PurePath
import fitz
from pdf4llm import to_markdown
from langchain.docstore.document import Document
from langchain.text_splitter import MarkdownTextSplitter
from .basepdf import BasePDF

class PDFMarkdownLoader(BasePDF):
    """
    Loader for PDF files converted content to markdown.
    """

    def __init__(
        self,
        path: PurePath,
        tokenizer: Callable[..., Any] = None,
        text_splitter: Callable[..., Any] = None,
        source_type: str = 'pdf',
        language: str = "eng",
        **kwargs
    ):
        super().__init__(
            path=path,
            tokenizer=tokenizer,
            text_splitter=text_splitter,
            source_type=source_type,
            language=language,
            **kwargs
        )
        self._splitter = MarkdownTextSplitter(chunk_size = 1024, chunk_overlap=10)

    def _load_pdf(self, path: Path) -> list:
        """
        Load a PDF file using the PDFMiner library.

        Args:
            path (Path): The path to the PDF file.

        Returns:
            list: A list of Langchain Documents.
        """
        if self._check_path(path):
            self.logger.info(f"Loading PDF file: {path}")
            docs = []
            pdf = fitz.open(str(path))
            md_text = to_markdown(pdf) # get markdown for all pages
            try:
                summary = self.get_summary_from_text(md_text)
            except Exception:
                summary = ''
            metadata = {
                "url": '',
                "filename": path.name,
                # "index": f"{path.name}",
                "source": str(path.name),
                "type": 'pdf',
                "question": '',
                "answer": '',
                "data": {},
                "summary": summary,
                "source_type": self._source_type,
                "document_meta": {
                    "title": pdf.metadata.get("title", ""),
                    # "subject": pdf.metadata.get("subject", ""),
                    # "keywords": pdf.metadata.get("keywords", ""),
                    "creationDate": pdf.metadata.get("creationDate", ""),
                    # "modDate": pdf.metadata.get("modDate", ""),
                    # "producer": pdf.metadata.get("producer", ""),
                    # "creator": pdf.metadata.get("creator", ""),
                    "author": pdf.metadata.get("author", ""),
                }
            }
            for idx, chunk in enumerate(self._splitter.split_text(md_text)):
                _info = {
                    "index": f"{idx}",
                    **metadata
                }
                docs.append(
                    Document(
                        page_content=chunk,
                        metadata=_info
                    )
                )
            return docs
        else:
            return []



from collections.abc import Callable
from typing import Any, Union, List
from abc import abstractmethod
from pathlib import Path
import subprocess
from .basevideo import BaseVideoLoader


class VideoLoader(BaseVideoLoader):
    """
    Generating Video transcripts from Videos.
    """
    _extension = ['.youtube']
    encoding = 'utf-8'
    chunk_size = 768

    def __init__(
        self,
        urls: List[str],
        tokenizer: Callable[..., Any] = None,
        text_splitter: Callable[..., Any] = None,
        source_type: str = 'video',
        language: str = "en",
        video_path: Union[str, Path] = None,
        **kwargs
    ):
        super().__init__(
            urls,
            tokenizer,
            text_splitter,
            source_type,
            language=language,
            video_path=video_path,
            **kwargs
        )

    def download_video(self, url: str, path: str) -> Path:
        """
        Downloads a video from a URL using yt-dlp.

        Args:
            video_url (str): The URL of the video to download.
            output_path (str): The directory where the video will be saved.
        """
        command = [
            "yt-dlp",
            "--get-filename",
            url
        ]
        try:
            result = subprocess.run(command, check=True, stdout=subprocess.PIPE, text=True)
            filename = result.stdout.strip()  # Remove any trailing newline characters
            file_path = path.joinpath(filename)
            if file_path.exists():
                print(f"Video already downloaded: {filename}")
                return file_path
            print(f"Downloading video: {filename}")
            # after extracted filename, download the video
            command = [
                "yt-dlp",
                url,
                "-o",
                str(file_path)
            ]
            subprocess.run(command, check=True)
            return file_path
        except subprocess.CalledProcessError as e:
            print(f"Error downloading video: {e}")


    def load(self) -> list:
        documents = []
        for url in self.urls:
            transcript = None
            if isinstance(url, dict):
                path = list(url.keys())[0]
                parts = url[path]
                if isinstance(parts, str):
                    video_title = parts
                elif isinstance(parts, dict):
                    video_title = parts['title']
                    transcript = parts.get('transcript', None)
                url = path
            else:
                video_title = url
            docs = self.load_video(url, video_title, transcript)
            documents.extend(docs)
        # return documents
        return self.split_documents(documents)

    @abstractmethod
    def load_video(self, url: str, video_title: str, transcript: str) -> list:
        pass

    def parse(self, source):
        pass



from langchain.chains.constitutional_ai.base import ConstitutionalChain
from langchain.chains.constitutional_ai.models import ConstitutionalPrinciple
from ...conf import ETHICAL_PRINCIPLE


ethical_principle = ConstitutionalPrinciple(
    name="Ethical Principle",
    critique_request=ETHICAL_PRINCIPLE,
    revision_request="Rewrite the model's output to be both ethical and legal.",
)


def get_constitutional_chain(llm, qa_chain):
    return ConstitutionalChain.from_llm(
        chain=qa_chain,
        constitutional_principles=[ethical_principle],
        llm=llm,
        verbose=True,
    )



from collections.abc import Callable
from pathlib import Path, PurePath
from typing import Union
import mammoth
import docx
from markdownify import markdownify as md
from langchain_community.document_loaders.word_document import (
    UnstructuredWordDocumentLoader
)
from langchain.text_splitter import MarkdownTextSplitter
from langchain.docstore.document import Document
from .abstract import AbstractLoader


class MSWordLoader(AbstractLoader):
    """
    Loader for Microsoft Word files.
    """
    _extension: list = ['.docx', '.doc']

    def __init__(
        self,
        path: PurePath,
        tokenizer: Union[str, Callable] = None,
        text_splitter: Union[str, Callable] = None,
        source_type: str = 'document',
        **kwargs
    ):
        self.path = path
        if isinstance(path, str):
            path = Path(path)
        self._use_mammoth: bool = kwargs.pop('use_mammoth', True)
        self._md_splitter = MarkdownTextSplitter(chunk_size = 1024, chunk_overlap=10)
        super().__init__(
            tokenizer=tokenizer,
            text_splitter=text_splitter,
            source_type=source_type,
            **kwargs
        )

    def _load_document(self, path: PurePath) -> list:
        if self._check_path(path):
            docs = []
            self.logger.info(f"Loading Word file: {path}")
            if self._use_mammoth:
                with open(path, "rb") as docx_file:
                    doc = docx.Document(str(path))
                    prop = doc.core_properties
                    result = mammoth.convert_to_html(docx_file)
                    # result = mammoth.extract_raw_text(docx_file)
                    html = result.value # The generated HTML
                    md_text = md(html) # The generated Markdown
                    try:
                        summary = self.get_summary_from_text(md_text)
                    except ValueError:
                        summary = ''
                    metadata = {
                        "url": '',
                        "source": path.name,
                        "filename": path.name,
                        "index": str(path.name),
                        "type": 'document',
                        "question": '',
                        "answer": '',
                        "source_type": self._source_type,
                        "data": {},
                        "summary": summary,
                        "document_meta": {
                            "author": prop.author,
                            "version": prop.version,
                            "title": prop.title,
                            "created": prop.created.strftime("%Y-%m-%d %H:%M:%S"),
                            "last_modified": prop.modified.strftime("%Y-%m-%d %H:%M:%S")
                        }
                    }
                    for chunk in self._md_splitter.split_text(md_text):
                        _idx = {
                            **metadata
                        }
                        docs.append(
                            Document(
                                page_content=chunk,
                                metadata=_idx
                            )
                        )
                return docs
            else:
                word_loader = UnstructuredWordDocumentLoader(
                    file_path=str(path)
                )
                docs = word_loader.load()
                for doc in docs:
                    # Fix This
                    doc.metadata['source_type'] = self._source_type
        return []

    def load(self) -> list:
        """
        Load data from a DOCX file.

        Args:
            source (str): The path to the DOCX file.

        Returns:
            list: A list of Langchain Documents.
        """
        if not self.path.exists():
            raise FileNotFoundError(f"DOCX file/directory not found: {self.path}")
        if self.path.is_dir():
            documents = []
            # iterate over the files in the directory
            for ext in self._extension:
                for item in self.path.glob(f'*{ext}'):
                    documents.extend(self._load_document(item))
        elif self.path.is_file():
            documents = self._load_document(self.path)
        else:
            raise ValueError(
                f"DOCX Loader: Invalid path: {self.path}"
            )
        # return documents
        return self.split_documents(documents)

    def parse(self, source):
        pass



# Langchain:
from langchain import hub
from ..exceptions import ConfigError  # pylint: disable=E0611

## LLM configuration
from .abstract import AbstractLLM

# Vertex
try:
    from .vertex import VertexLLM
    VERTEX_ENABLED = True
except (ModuleNotFoundError, ImportError):
    VERTEX_ENABLED = False

# Anthropic:
try:
    from .anthropic import Anthropic
    ANTHROPIC_ENABLED = True
except (ModuleNotFoundError, ImportError):
    ANTHROPIC_ENABLED = False

# OpenAI
try:
    from .openai import OpenAILLM
    OPENAI_ENABLED = True
except (ModuleNotFoundError, ImportError):
    OPENAI_ENABLED = False

# LLM Transformers
try:
    from  .pipes import PipelineLLM
    TRANSFORMERS_ENABLED = True
except (ModuleNotFoundError, ImportError):
    TRANSFORMERS_ENABLED = False

# HuggingFaces Hub:
try:
    from  .hf import HuggingFace
    HF_ENABLED = True
except (ModuleNotFoundError, ImportError):
    HF_ENABLED = False

# GroQ:
try:
    from .groq import GroqLLM
    GROQ_ENABLED = True
except (ModuleNotFoundError, ImportError):
    GROQ_ENABLED = False

# Mixtral:
try:
    from .groq import GroqLLM
    MIXTRAL_ENABLED = True
except (ModuleNotFoundError, ImportError):
    MIXTRAL_ENABLED = False

# Google
try:
    from .google import GoogleGenAI
    GOOGLE_ENABLED = True
except (ModuleNotFoundError, ImportError):
    GOOGLE_ENABLED = False


def get_llm(llm_name: str, model_name: str, **kwargs) -> AbstractLLM:
    """get_llm.

    Load the Language Model for the Chatbot.
    """
    print('LLM > ', llm_name)
    if llm_name == 'VertexLLM':
        if VERTEX_ENABLED is False:
            raise ConfigError(
                "VertexAI enabled but not installed."
            )
        return VertexLLM(model=model_name, **kwargs)
    if llm_name == 'Anthropic':
        if ANTHROPIC_ENABLED is False:
            raise ConfigError(
                "ANTHROPIC is enabled but not installed."
            )
        return Anthropic(model=model_name, **kwargs)
    if llm_name == 'OpenAI':
        if OPENAI_ENABLED is False:
            raise ConfigError(
                "OpenAI is enabled but not installed."
            )
        return OpenAILLM(model=model_name, **kwargs)
    if llm_name == 'hf':
        if HF_ENABLED is False:
            raise ConfigError(
                "Hugginfaces Hub is enabled but not installed."
            )
        return HuggingFace(model=model_name, **kwargs)
    if llm_name == 'pipe':
        if TRANSFORMERS_ENABLED is False:
            raise ConfigError(
                "Transformes Pipelines are enabled, but not installed."
            )
        return PipelineLLM(model=model_name, **kwargs)
    if llm_name in ('Groq', 'llama3', 'mixtral'):
        if GROQ_ENABLED is False:
            raise ConfigError(
                "Groq is enabled but not installed."
            )
        return GroqLLM(model=model_name, **kwargs)
    if llm_name == 'Google':
        if GOOGLE_ENABLED is False:
            raise ConfigError(
                "Google is enabled but not installed."
            )
        return GoogleGenAI(model=model_name, **kwargs)
    else:
        # TODO: Add more LLMs
        return hub.pull(llm_name)

def get_llm_list():
    """get_llm_list.

    Get the list of available LLMs.
    """
    llms = []
    if VERTEX_ENABLED:
        llms.append(VertexLLM.get_supported_models())
    if ANTHROPIC_ENABLED:
        llms.append(Anthropic.get_supported_models())
    if OPENAI_ENABLED:
        llms.append(OpenAILLM.get_supported_models())
    if HF_ENABLED:
        llms.append(HuggingFace.get_supported_models())
    if TRANSFORMERS_ENABLED:
        llms.append(PipelineLLM.get_supported_models())
    if GROQ_ENABLED:
        llms.append(GroqLLM.get_supported_models())
    if GOOGLE_ENABLED:
        llms.append(GoogleGenAI.get_supported_models())
    return llms



from langchain_experimental.tools.python.tool import PythonAstREPLTool
from .base import BaseAgent
# importing Tools
from ..tools import (
    ZipcodeAPIToolkit,
    WikipediaTool,
    WikidataTool,
    GoogleSearchTool,
    GoogleLocationFinder,
    BingSearchTool,
    AskNewsTool,
    DuckDuckGoSearchTool,
    YouTubeSearchTool,
    OpenWeatherMapTool,
    StackExchangeTool,
)


class CopilotAgent(BaseAgent):
    """CopilotAgent Agent.

    This is Agent Base class for AI Copilots.
    """
    def __init__(
        self,
        name: str = 'Agent',
        llm: str = 'vertexai',
        tools: list = None,
        prompt_template: str = None,
        **kwargs
    ):
        super().__init__(name, llm, tools, prompt_template, **kwargs)
        self.tools = [
                PythonAstREPLTool(
                    name='python_repl_ast',
                    globals={},
                    locals={}
                )
            ] + list(tools)
        self.prompt = self.get_prompt(
            self.prompt_template
        )
        # print('PROMPT > ', self.prompt)

    @classmethod
    def default_tools(cls) -> list:
        # ZipCode API Toolkit
        tools = []
        zpt_tools = []
        try:
            zpt = ZipcodeAPIToolkit()
            zpt_tools = zpt.get_tools()

        except Exception as e:
            print('ERROR LOADING ZIPCODE TOOLS')

        try:
            wk1 = WikipediaTool()
            wk12 = WikidataTool()

            g1 = GoogleSearchTool()
            g2 = GoogleLocationFinder()

            b = BingSearchTool()
            d = DuckDuckGoSearchTool()
            ask = AskNewsTool()

            yt = YouTubeSearchTool()
            stackexchange = StackExchangeTool()
            weather = OpenWeatherMapTool()

            tools = [
                wk1,
                wk12, g1, g2,
                b, d, ask, yt,
                weather,
                stackexchange
            ] + zpt_tools

        except Exception as e:
            print('TOOL Error > ', e)

        return tools



from typing import Union
from pathlib import Path
from pydantic import PrivateAttr
from crewai_tools import BaseTool, RagTool

class RagSearchTool(BaseTool):
    """RagTool is designed to answer questions by leveraging the power of RAG by leveraging (EmbedChain)."""
    name: str = "RAG Tool"
    description: str = "enables users to dynamically query a knowledge base"
    _directory: PrivateAttr

    def __init__(self, directory: Union[Path, str] = None, **kwargs):
        super().__init__(**kwargs)
        self._directory = directory

    def get_rag(self):
        """Return the RAG Tool."""
        return RagTool(
            config={
                "llm": {
                    "provider": "vertexai",
                    "config": {
                        "model": "gemini-pro",
                        "temperature": 0.4,
                        "top_p": 1,
                        "stream": True
                    }
                },
                "embedder": {
                    "provider": "vertexai",
                    "config": {
                        "model": "embedding-001"
                    }
                }
            }
        )

    def _run(self, query: str, **kwargs) -> dict:
        """Query Several sources of information for knowledge base."""
        rag = self.get_rag()
        if self._directory:
            return rag.from_directory(self._directory)



from typing import Optional, Dict, Any, Type
from abc import abstractmethod
from langchain_core.pydantic_v1 import BaseModel, Field, Extra
from langchain_core.callbacks import CallbackManagerForToolRun
from langchain_core.tools import BaseTool

class AbstractToolArgsSchema(BaseModel):
    """Schema for the arguments to the AbstractTool."""

    # This Field allows any number of arguments to be passed in.
    args: list = Field(description="A list of arguments to the tool")


class AbstractTool(BaseTool):
    """Abstract class for tools."""

    args_schema: Type[BaseModel] = AbstractToolArgsSchema

    class Config:
        """Configuration for this pydantic object."""
        extra = Extra.forbid
        arbitrary_types_allowed = True

    @abstractmethod
    def _search(self, query: str) -> str:
        """Run the tool."""

    async def _asearch(self, *args, **kwargs):
        """Run the tool asynchronously."""
        return self._search(*args, **kwargs)

    def _run(
        self,
        query: str,
        run_manager: Optional[CallbackManagerForToolRun] = None
    ) -> Dict[str, Any]:
        args = [a.strip() for a in query.split(',')]
        try:
            return self._search(*args)
        except Exception as e:
            raise ValueError(f"Error running tool: {e}") from e

    async def _arun(
        self,
        query: str,
        run_manager: Optional[CallbackManagerForToolRun] = None
    ) -> Dict[str, Any]:
        """Use the tool asynchronously."""
        args = [a.strip() for a in query.split(',')]
        try:
            return await self._asearch(*args)
        except Exception as e:
            raise ValueError(f"Error running tool: {e}") from e



from collections.abc import Callable
from typing import Any
import uuid
import asyncio
from aiohttp import web
from sentence_transformers import SentenceTransformer, util
from langchain.memory import (
    ConversationSummaryMemory,
    ConversationBufferMemory
)
from langchain.chains.retrieval_qa.base import RetrievalQA
from langchain.chains.conversational_retrieval.base import (
    ConversationalRetrievalChain
)
from langchain.retrievers import (
    EnsembleRetriever,
    ContextualCompressionRetriever
)
from langchain_core.vectorstores import VectorStoreRetriever
from langchain_core.prompts import (
    PromptTemplate,
    ChatPromptTemplate
)
from langchain_community.retrievers import BM25Retriever
from datamodel.exceptions import ValidationError  # pylint: disable=E0611
from asyncdb import AsyncDB
from navconfig.logging import logging
from navigator_session import get_session
from parrot.conf import (
    BIGQUERY_CREDENTIALS,
    BIGQUERY_PROJECT_ID,
    BIGQUERY_DATASET
)
try:
    from ...llms import VertexLLM
    VERTEX_ENABLED = True
except ImportError:
    VERTEX_ENABLED = False

try:
    from ...llms import Anthropic
    ANTHROPIC_ENABLED = True
except ImportError:
    ANTHROPIC_ENABLED = False

from ...utils import SafeDict
from ...models import ChatResponse, ChatbotUsage


class RetrievalManager:
    """Managing the Chain Retrieval, answers and sources.
    """
    def __init__(
        self,
        chatbot_id: uuid.UUID,
        chatbot_name: str,
        model: Callable,
        store: Callable,
        memory: ConversationBufferMemory = None,
        template: str = None,
        source_path: str = 'web',
        request: web.Request = None,
        kb: Any = None
    ):
        # Chatbot ID:
        self.chatbot_id: uuid.UUID = chatbot_id
        # Chatbot Name:
        self.chatbot_name: str = chatbot_name
        # Source Path:
        self.source_path: str = source_path
        # Vector Store
        self.store = store
        # Memory Manager
        self.memory = memory
        # LLM Model
        self.model = model
        # template prompt
        # TODO: if none, create a basic template
        self.template = template
        # Knowledge-base
        self.kb = kb
        # Logger:
        self.logger = logging.getLogger('Parrot.Retrieval')
        # Web Request:
        self.request = request

    def __enter__(self):
        self.client_id: str = str(uuid.uuid4())
        self.client, _ = self.store.connect(alias=self.client_id)
        return self

    async def __aenter__(self):
        self.client_id: str = str(uuid.uuid4())
        self.client, _ = self.store.connect(alias=self.client_id)
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        self.client.close()  # closing database connection
        # closing the connection:
        self.store.close(alias=self.client_id)
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        self.client.close()  # closing database connection
        # closing the connection:
        self.store.close(alias=self.client_id)
        pass

    def create_memory(
        self,
        key: str = 'chat_history',
        input_key: str = 'question',
        output_key: str = 'answer'
    ):
        return ConversationBufferMemory(
            memory_key=key,
            return_messages=True,
            input_key=input_key,
            output_key=output_key
        )

    ### Different types of Retrieval
    def conversation(
        self,
        question: str = None,
        chain_type: str = 'stuff',
        search_type: str = 'similarity',
        search_kwargs: dict = {"k": 4, "fetch_k": 10, "lambda_mult": 0.89},
        return_docs: bool = True,
        metric_type: str = None,
        memory: Any = None,
        use_llm: str = None,
        **kwargs
    ):
        # Question:
        self._question = question
        # Memory:
        self.memory = memory
        # Get a Vector Retriever:
        vector = self.store.get_vector(
            metric_type=metric_type
        )
        simil_retriever = VectorStoreRetriever(
            vectorstore=vector,
            search_type='similarity',
            chain_type=chain_type,
            search_kwargs=search_kwargs
        )
        retriever = vector.as_retriever(
            search_type=search_type,
            search_kwargs=search_kwargs
        )
        if self.kb:
            # Get a BM25 Retriever:
            b25_retriever = BM25Retriever.from_documents(self.kb)
            retriever = EnsembleRetriever(
                retrievers=[simil_retriever, retriever, b25_retriever],
                weights=[0.6, 0.3, 0.1]
            )
        else:
            retriever = EnsembleRetriever(
                retrievers=[simil_retriever, retriever],
                weights=[0.7, 0.3]
            )
        custom_template = self.template.format_map(
            SafeDict(
                summaries=''
            )
        )
        _prompt = PromptTemplate(
            template=custom_template,
            input_variables=["context", "chat_history", "question"],
        )
        if use_llm is not None:
            if use_llm == 'claude':
                if ANTHROPIC_ENABLED is True:
                    llm = Anthropic(
                        model='claude-3-opus-20240229',
                        temperature=0.2,
                        top_p=0.4,
                        top_k=20
                    )
                else:
                    raise ValueError(
                        "No Anthropic Claude was installed."
                    )
            elif use_llm == 'vertex':
                if VERTEX_ENABLED is True:
                    llm = VertexLLM(
                        model='gemini-pro',
                        temperature=0.2,
                        top_p=0.4,
                        top_k=20
                    )
                else:
                    raise ValueError(
                        "No VertexAI was installed."
                    )
            else:
                raise ValueError(
                    f"Only Claude and Vertex are Supported Now."
                )
            _model = llm.get_llm()
        else:
            _model = self.model
        # Conversational Chain:
        self.chain = ConversationalRetrievalChain.from_llm(
            llm=_model,
            retriever=retriever,
            chain_type=chain_type,
            verbose=True,
            memory=self.memory,
            return_source_documents=return_docs,
            return_generated_question=True,
            combine_docs_chain_kwargs={"prompt": _prompt},
        )
        return self

    def qa(
        self,
        question: str = None,
        chain_type: str = 'stuff',
        search_type: str = 'mmr',
        search_kwargs: dict = {"k": 4, "fetch_k": 10, "lambda_mult": 0.89},
        return_docs: bool = True,
        metric_type: str = None,
        use_llm: str = None
    ):
        # Question:
        self._question = question
        # Get a Vector Retriever:
        vector = self.store.get_vector(
            metric_type=metric_type
        )
        simil_retriever = VectorStoreRetriever(
            vectorstore=vector,
            search_type='similarity',
            chain_type=chain_type,
            search_kwargs=search_kwargs
        )
        retriever = vector.as_retriever(
            search_type=search_type,
            search_kwargs=search_kwargs
        )
        if self.kb:
            # Get a BM25 Retriever:
            b25_retriever = BM25Retriever.from_documents(self.kb)
            retriever = EnsembleRetriever(
                retrievers=[simil_retriever, retriever, b25_retriever],
                weights=[0.6, 0.3, 0.1]
            )
        else:
            retriever = EnsembleRetriever(
                retrievers=[simil_retriever, retriever],
                weights=[0.7, 0.3]
            )
            # retriever = simil_retriever
        custom_template = self.template.format_map(
            SafeDict(
                chat_history=''
            )
        )
        if use_llm is not None:
            if use_llm == 'claude':
                if ANTHROPIC_ENABLED is True:
                    llm = Anthropic(
                        model='claude-3-opus-20240229',
                        temperature=0.2,
                        top_p=0.4,
                        top_k=20
                    )
                else:
                    raise ValueError(
                        "No Anthropic Claude was installed."
                    )
            elif use_llm == 'vertex':
                if VERTEX_ENABLED is True:
                    llm = VertexLLM(
                        model='gemini-pro',
                        temperature=0.2,
                        top_p=0.4,
                        top_k=20
                    )
                else:
                    raise ValueError(
                        "No VertexAI was installed."
                    )
            else:
                raise ValueError(
                    f"Only Claude and Vertex are Supported Now."
                )
            self.model = llm.get_llm()

        self.chain = RetrievalQA.from_chain_type(
            llm=self.model,
            chain_type=chain_type,
            retriever=retriever,
            return_source_documents=return_docs,
            verbose=True,
            chain_type_kwargs={
                "prompt": PromptTemplate(
                    template=custom_template,
                    input_variables=['context', 'question']
                )
            },
        )
        # Debug Code ::
        # print('=====================')
        # print(custom_template)
        # response = self.chain.invoke(question)
        # print('Q > ', response['result'])
        # docs = vector.similarity_search(
        #     self._question, k=10
        # )
        # print(" LENGHT DOCS > ", len(docs))
        # print(docs)
        # print(' ========================== ')

        # try:
        #     distance = self.evaluate_distance(
        #         self.store.embedding_name, question, docs
        #     )
        #     print('DISTANCE > ', distance)
        # except Exception as e:
        #     distance = 'EMPTY'
        # print('DISTANCE > ', distance)
        # print('CHAIN > ', self.chain)

        return self

    def get_current_context(self):
        if self.memory:
            return self.memory.buffer_as_str()
        return None

    def as_markdown(self, response: ChatResponse) -> str:
        markdown_output = f"**Question**: {response.question}  \n"
        markdown_output += f"**Answer**: {response.answer}  \n"
        if response.source_documents:
            source_documents = response.source_documents
            current_sources = []
            block_sources = []
            count = 0
            d = {}
            for source in source_documents:
                if count >= 20:
                    break  # Exit loop after processing 10 documents
                metadata = source.metadata
                if 'url' in metadata:
                    src = metadata.get('url')
                elif 'filename' in metadata:
                    src = metadata.get('filename')
                else:
                    src = metadata.get('source', 'unknown')
                if src == 'knowledge-base':
                    continue  # avoid attaching kb documents
                source_title = metadata.get('title', src)
                if source_title in current_sources:
                    continue
                current_sources.append(source_title)
                if src:
                    d[src] = metadata.get('document_meta', {})
                source_filename = metadata.get('filename', src)
                if src:
                    block_sources.append(f"- [{source_title}]({src})")
                else:
                    if 'page_number' in metadata:
                        block_sources.append(f"- {source_filename} (Page {metadata.get('page_number')})")
                    else:
                        block_sources.append(f"- {source_filename}")
            if block_sources:
                markdown_output += f"**Sources**:  \n"
                markdown_output += "\n".join(block_sources)
            if d:
                response.documents = d
        return markdown_output

    def evaluate_distance(self, model, question, source_documents):
        tokenizer = SentenceTransformer(model)
        query_embedding = tokenizer.encode(question)
        document_embeddings = [
            tokenizer.encode(doc.page_content) for doc in source_documents
        ]
        distances = util.cos_sim(query_embedding, document_embeddings)
        result = []
        for doc, distance in zip(source_documents, distances):
            result.append({
                "document": doc,
                "distance": distance
            })
        return result

    async def log_usage(self, response: ChatResponse, request: web.Request = None):
        PARAMS = {
            "credentials": BIGQUERY_CREDENTIALS,
            "project_id": BIGQUERY_PROJECT_ID,
        }
        db = AsyncDB(
            'bigquery',
            params=PARAMS
        )
        origin = {
            "user_agent": 'script'
        }
        user_id = 0
        if request:
            origin = {
                "origin": request.remote,
                "user_agent": request.headers.get('User-Agent')
            }
            session = await get_session(request)
            if session:
                user_id = session.user_id
        async with await db.connection() as conn:  #pylint: disable=E1101
            # set connection to model:
            ChatbotUsage.Meta.connection = conn
            # Add a new record of chatbot usage:
            record = {
                "chatbot_id": str(self.chatbot_id),
                "user_id": user_id,  # TODO: add session informtion
                "source_path": self.source_path,
                "platform": 'web',
                "sid": str(response.sid),
                "used_at": response.at,
                "question": response.question,
                "response": response.answer,
                **origin
            }
            try:
                log = ChatbotUsage(**record)
                data = log.to_dict()
                # convert to string (bigquery uses json.dumps to convert to string)
                data['sid'] = str(data['sid'])
                data['chatbot_id'] = str(data['chatbot_id'])
                data['event_timestamp'] = str(data['event_timestamp'])
                # writing directly to bigquery
                await conn.write(
                    [data],
                    table_id=ChatbotUsage.Meta.name,
                    dataset_id=ChatbotUsage.Meta.schema,
                    use_streams=False,
                    use_pandas=False
                )
                # await log.insert()
            except Exception as exc:
                self.logger.error(
                    f"Error inserting log: {exc}"
                )


    async def question(
            self,
            question: str = None,
            chain_type: str = 'stuff',
            search_type: str = 'similarity',
            search_kwargs: dict = {"k": 4, "fetch_k": 10, "lambda_mult": 0.89},
            return_docs: bool = True,
            metric_type: str = None,
            memory: Any = None,
            **kwargs
    ):
        # Generating Vector:
        async with self.store.connection() as store:  #pylint: disable=E1101
            vector = store.get_vector(metric_type=metric_type)
            retriever = VectorStoreRetriever(
                vectorstore=vector,
                search_type=search_type,
                chain_type=chain_type,
                search_kwargs=search_kwargs
            )
            custom_template = self.template.format_map(
                SafeDict(
                    summaries=''
                )
            )
            _prompt = PromptTemplate(
                template=custom_template,
                input_variables=["context", "chat_history", "question"],
            )
            try:
                chain = ConversationalRetrievalChain.from_llm(
                    llm=self.model,
                    retriever=retriever,
                    chain_type=chain_type,
                    verbose=False,
                    memory=memory,
                    return_source_documents=return_docs,
                    return_generated_question=True,
                    combine_docs_chain_kwargs={"prompt": _prompt},
                    **kwargs
                )
                response = chain.invoke(
                    question
                )
            except Exception as exc:
                self.logger.error(
                    f"Error invoking chain: {exc}"
                )
                return None, None
            try:
                qa_response = ChatResponse(**response)
            except (ValueError, TypeError) as exc:
                self.logger.error(
                    f"Error validating response: {exc}"
                )
                return None, None
            except ValidationError as exc:
                self.logger.error(
                    f"Error on response: {exc.payload}"
                )
                return None, None
        try:
            qa_response.response = self.as_markdown(
                qa_response
            )
            # saving question to Usage Log
            if self.request:
                tasker = self.request.app['service_queue']
                await tasker.put(
                    self.log_usage,
                    response=qa_response,
                    request=self.request
                )
            else:
                asyncio.create_task(
                    self.log_usage(response=qa_response)
                )
            return qa_response
        except Exception as exc:
            self.logger.exception(
                f"Error on response: {exc}"
            )
            return None


    async def invoke(self, question):
        # Invoke the chain with the given question
        try:
            response = self.chain.invoke(
                question
            )
        except Exception as exc:
            self.logger.error(
                f"Error invoking chain: {exc}"
            )
            return None, None
        try:
            qa_response = ChatResponse(**response)
        except (ValueError, TypeError) as exc:
            self.logger.error(
                f"Error validating response: {exc}"
            )
            return None, None
        except ValidationError as exc:
            self.logger.error(
                f"Error on response: {exc.payload}"
            )
            return None, None
        try:
            qa_response.response = self.as_markdown(
                qa_response
            )
            # saving question to Usage Log
            if self.request:
                tasker = self.request.app['service_queue']
                await tasker.put(
                    self.log_usage,
                    response=qa_response,
                    request=self.request
                )
            else:
                asyncio.create_task(self.log_usage(response=qa_response))
            return qa_response
        except Exception as exc:
            self.logger.exception(
                f"Error on response: {exc}"
            )
            return None, None



from collections.abc import Callable
from typing import Any, Optional, List, Union
from pathlib import PurePath, Path
from io import StringIO
import fitz  # PyMuPDF
from langchain.docstore.document import Document
from langchain.text_splitter import (
    RecursiveCharacterTextSplitter
)
from .basepdf import BasePDF


class PDFChapterLoader(BasePDF):
    """
    Preserving Chapter Structure from PDF files.
    """
    def __init__(
        self,
        path: PurePath,
        tokenizer: Callable[..., Any] = None,
        text_splitter: Callable[..., Any] = None,
        source_type: str = 'pdf',
        language: str = "eng",
        **kwargs
    ):
        super().__init__(
            path=path,
            tokenizer=tokenizer,
            text_splitter=text_splitter,
            source_type=source_type,
            language=language,
            **kwargs
        )
        # Which Font is used for titles (Chapter separation)
        self.title_font: list = kwargs.get('title_font', 'Calibri-Bold')
        if not text_splitter:
            self.text_splitter = RecursiveCharacterTextSplitter.from_huggingface_tokenizer(
                self.tokenizer,
                chunk_size=2000,
                chunk_overlap=100,
                add_start_index=True,  # If `True`, includes chunk's start index in metadata
                strip_whitespace=True,  # strips whitespace from the start and end
                separators=["\n\n", "\n", "\r\n", "\r", "\f", "\v", "\x0b", "\x0c"],
            )

    def eval_title(self, title_font: str) -> bool:
        """
        Check if the font is a title font.

        Args:
            title_font (str): The font to check.

        Returns:
            bool: True if the font is a title font.
        """
        return 'Bold' in title_font or title_font == self.title_font

    def _load_pdf(self, path: PurePath, **kwargs):
        """
        Open a PDF file.

        Args:
            path (PurePath): The path to the PDF file.

        Returns:
            pdfplumber.PDF: The PDF object.
        """
        pdf = fitz.open(path)
        self.logger.info(f"Loading PDF file: {path}")
        chapters = []
        current_chapter_text = ''
        current_chapter_title = ''
        current_chapter_page = None
        chapter_titles = set()  # Keep track of unique chapter titles
        for page_num in range(len(pdf)):
            page = pdf.load_page(page_num)
            blocks = page.get_text("dict")["blocks"]
            page_number = page_num + 1
            metadata = {
                "url": '',
                "index": f"{path.name} #{page_number}",
                "source": f"{path.name} #{page_number}",
                "filename": path.name,
                "source_type": self._source_type,
                "type": "pdf",
                "question": "",
                "answer": "",
                "summary": '',
                "document_meta": {
                    "page_number": page_num,
                    # **pdf.metadata
                }
            }
            for b in blocks:
                if b['type'] == 0:  # Text block
                    block_text = ''
                    for line in b["lines"]:
                        for span in line["spans"]:
                            block_text += span['text']  # Accumulate text within the block

                    # Check if the block text is a title by examining the font
                    if any(self.eval_title(span['font']) for line in b["lines"] for span in line["spans"]):
                        title = block_text.strip()
                        if title not in chapter_titles:
                            # Save the current chapter if it's not empty and start a new one
                            if current_chapter_text.strip() and current_chapter_text.strip() != current_chapter_title.strip():
                                chapters.append({
                                    'chapter': current_chapter_title,
                                    'content': current_chapter_text.strip(),
                                    'page': current_chapter_page,
                                    'meta': metadata
                                })
                            current_chapter_title = f"**{title}**: "
                            current_chapter_page = page_num + 1
                            current_chapter_text = current_chapter_title
                            chapter_titles.add(title)
                        else:
                            # Continue appending to the existing chapter
                            current_chapter_text += block_text
                    else:
                        # Continue appending text to the current chapter
                        current_chapter_text += block_text

            # Add a newline after processing each block, if not a chapter title
            if not block_text.strip().startswith(current_chapter_title):
                current_chapter_text += "\n"

        # Save the last chapter if it exists and it's not just the title
        if current_chapter_text.strip() and current_chapter_text.strip() != current_chapter_title.strip():
            chapters.append({
                'chapter': current_chapter_title,
                'content': current_chapter_text.strip(),
                'page': current_chapter_page,
                'meta': metadata
            })
        documents = []
        for chapter in chapters:
            documents.append(Document(
                page_content=chapter['content'],
                metadata=chapter['meta']
            ))
        return documents



# -*- coding: utf-8 -*-
"""Navigator Parrot.

Basic Chatbots for Navigator Services.
"""
import os
from pathlib import Path
from .version import (
    __author__,
    __author_email__,
    __description__,
    __title__,
    __version__
)

os.environ["USER_AGENT"] = "Parrot.AI/1.0"

def get_project_root() -> Path:
    return Path(__file__).parent.parent

ABS_PATH = get_project_root()



from typing import Optional, Union
from datetime import datetime
from pathlib import PurePath
from datamodel import BaseModel, Field

def uuid_to_str(obj) -> str:
    return str(obj)

class BotData(BaseModel):
    chatbot_id: str = Field(primary_key=True, required=True, encoder=uuid_to_str)
    name: str = Field(required=True)
    source_type: str = Field(required=True, default='content')
    category: str = Field(required=True, default='data')
    tags: Optional[list[str]] = Field(required=False, default_factory=list)
    document_type: str = Field(required=False, default='document')
    loader: str = Field(required=True, default='TXTLoader')
    source_path: Union[str,PurePath] = Field(required=False)
    extensions: list[str] = Field(required=False)
    data: Optional[Union[list,dict]] = Field(required=False)
    arguments: Optional[dict] = Field(default_factory=dict)
    version: int = Field(required=True, default=1)
    updated_at: datetime = Field(required=False, default=datetime.now)

    class Meta:
        name: str = 'chatbots_data'



#!/usr/bin/env python
"""Navigator AI Parrot.

    Chatbot services for Navigator, based on Langchain.
See:
https://github.com/phenobarbital/ai-parrot
"""
import ast
from os import path

from Cython.Build import cythonize
from setuptools import Extension, find_packages, setup
from setuptools.command.build_ext import build_ext

def get_path(filename):
    return path.join(path.dirname(path.abspath(__file__)), filename)


def readme():
    with open(get_path('README.md'), encoding='utf-8') as rd:
        return rd.read()


version = get_path('parrot/version.py')
with open(version, 'r', encoding='utf-8') as meta:
    # exec(meta.read())
    t = compile(meta.read(), version, 'exec', ast.PyCF_ONLY_AST)
    for node in (n for n in t.body if isinstance(n, ast.Assign)):
        if len(node.targets) == 1:
            name = node.targets[0]
            if isinstance(name, ast.Name) and \
                    name.id in {
                        '__version__',
                        '__title__',
                        '__description__',
                        '__author__',
                        '__license__', '__author_email__'}:
                v = node.value
                if name.id == '__version__':
                    __version__ = v.s
                if name.id == '__title__':
                    __title__ = v.s
                if name.id == '__description__':
                    __description__ = v.s
                if name.id == '__license__':
                    __license__ = v.s
                if name.id == '__author__':
                    __author__ = v.s
                if name.id == '__author_email__':
                    __author_email__ = v.s

COMPILE_ARGS = ["-O2"]

extensions = [
    Extension(
        name='parrot.exceptions',
        sources=['parrot/exceptions.pyx'],
        extra_compile_args=COMPILE_ARGS,
        language="c"
    ),
    Extension(
        name='parrot.utils.types',
        sources=['parrot/utils/types.pyx'],
        extra_compile_args=COMPILE_ARGS,
        language="c++"
    ),
    Extension(
        name='parrot.utils.parsers.toml',
        sources=['parrot/utils/parsers/toml.pyx'],
        extra_compile_args=COMPILE_ARGS,
        language="c"
    ),
]

# Custom build_ext command to ensure cythonization during the build step
class BuildExtensions(build_ext):
    """Custom build_ext command to ensure cythonization during the build step."""
    def build_extensions(self):
        try:
            from Cython.Build import cythonize  # pylint: disable=import-outside-toplevel
            self.extensions = cythonize(self.extensions)
        except ImportError:
            print(
                "Cython not found. Extensions will be compiled without cythonization!"
            )
        super().build_extensions()

setup(
    name=__title__,
    version=__version__,
    author='Jesus Lara',
    author_email='jesuslara@phenobarbital.info',
    url='https://github.com/phenobarbital/ai-parrot',
    description=__description__,
    long_description=readme(),
    long_description_content_type='text/markdown',
    license=__license__,
    python_requires=">=3.10.12",
    keywords=['asyncio', 'asyncpg', 'aioredis', 'aiomcache', 'langchain', 'chatbot', 'agents'],
    platforms=['POSIX'],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Operating System :: POSIX :: Linux",
        "Environment :: Web Environment",
        "License :: OSI Approved :: MIT License",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3 :: Only",
        "Framework :: AsyncIO",
    ],
    packages=find_packages(
        exclude=[
            "bin",
            "contrib",
            "docs",
            "documents",
            "tests",
            "examples",
            "libraries",
            "db",
            "cache",
            ".jupyter",
            "locale",
            "lab",
            "notebooks",
            "resources",
            "static",
            "templates"
        ]
    ),
    package_data={"parrot": ["py.typed"]},
    setup_requires=[
        "setuptools_scm>=8.0.4",
        "wheel>=0.44.0",
        'Cython==3.0.11',
    ],
    install_requires=[
        "Cython==3.0.11",
        "accelerate==0.34.2",
        "langchain>=0.2.6",
        "langchain-community>=0.2.6",
        "langchain-core>=0.2.32",
        "langchain-experimental==0.0.62",
        "langchainhub==0.1.15",
        "langchain-text-splitters==0.2.2",
        "langchain-huggingface==0.0.3",
        "huggingface-hub==0.23.5",
        "llama-index==0.10.20",
        "llama_cpp_python==0.2.56",
        "bitsandbytes==0.43.3",
        "Cartopy==0.22.0",
        "chromadb==0.4.24",
        "datasets==2.18.0",
        "faiss-cpu==1.8.0",
        "fastavro==1.9.4",
        "gunicorn==21.2.0",
        "jq==1.7.0",
        "rank_bm25==0.2.2",
        "matplotlib==3.8.3",
        "numba==0.59.0",
        "querysource>=3.12.10",
        "safetensors>=0.4.3",
        "sentence-transformers==3.0.1",
        "tabulate==0.9.0",
        "tiktoken==0.7.0",
        "tokenizers==0.19.1",
        "unstructured==0.14.3",
        "unstructured-client==0.18.0",
        "youtube-transcript-api==0.6.2",
        "selenium==4.18.1",
        "webdriver_manager==4.0.1",
        "transitions==0.9.0",
        "sentencepiece==0.2.0",
        "duckduckgo-search==5.3.0",
        "google-search-results==2.4.2",
        "google-api-python-client>=2.86.0",
        "gdown==5.1.0",
        "weasyprint==61.2",
        "markdown2==2.4.13",
        "fastembed==0.3.4",
        "yfinance==0.2.40",
        "youtube_search==2.1.2",
        "wikipedia==1.4.0",
        "mediawikiapi==1.2",
        # "wikibase-rest-api-client==0.2.1",
        # "asknews==0.7.30",
        "pyowm==3.3.0",
        "O365==2.0.35",
        "stackapi==0.3.1",
        "timm==1.0.9",
        "torchvision==0.19.1",
    ],
    extras_require={
        "loaders": [
            "pymupdf==1.24.4",
            "pymupdf4llm==0.0.1",
            "pdf4llm==0.0.6",
            "PyPDF2==3.0.1",
            "pdfminer.six==20231228",
            "pdfplumber==0.11.0",
            "GitPython==3.1.42",
            "opentelemetry-sdk==1.24.0",
            "rapidocr-onnxruntime==1.3.15",
            "pytesseract==0.3.10",
            "python-docx==1.1.0",
            "python-pptx==0.6.23",
            "docx2txt==0.8",
            "pytube==15.0.0",
            "pydub==0.25.1",
            "markdownify==0.12.1",
            "yt_dlp==2024.4.9",
            "moviepy==1.0.3",
            "mammoth==1.7.1",
            "paddlepaddle==2.6.1",
            "paddlepaddle_gpu==2.6.1",
            "paddleocr==2.8.1",
            "ftfy==6.2.3",
            "librosa==0.10.1",
            "XlsxWriter==3.2.0",
            "xformers==0.0.27.post2",
        ],
        "anthropic": [
            "langchain-anthropic==0.1.11",
            "anthropic==0.25.2",
        ],
        "openai": [
            "langchain-openai==0.1.21",
            "openai==1.40.3",
            "llama-index-llms-openai==0.1.11",
            "tiktoken==0.7.0"
        ],
        "google": [
            "langchain-google-vertexai==1.0.10",
            "langchain-google-genai==1.0.10",
            "vertexai==1.65.0"
        ],
        "hunggingfaces": [
            "llama-index-llms-huggingface==0.2.7"
        ],
        "groq": [
            "groq==0.11.0",
            "langchain-groq==0.1.9"
        ],
        "qdrant": [
            "qdrant-client==1.8.0",
        ],
        "milvus": [
            "langchain-milvus>=0.1.4",
            "milvus==2.3.5",
            "pymilvus==2.4.6",
        ],
        "crew": [
            "colbert-ai==0.2.19",
            "vanna==0.3.4", # Vanna:
            "crewai[tools]==0.28.8"
        ],
        "analytics": [
            "annoy==1.17.3",
            "gradio_tools==0.0.9",
            "gradio-client==0.2.9",
            "streamlit==1.37.1",
            "simsimd==4.3.1",
            "opencv-python==4.10.0.84"
            # "timm==0.9.16", # image-processor
            # "ultralytics==8.2.4", # image-processor
            # "albumentations-1.4.4",
            # "yolov8-0.0.2"
        ],
        "all": [
            # Install all LLM Models
            "langchain-milvus==0.1.1",
            "milvus==2.3.5",
            "pymilvus==2.4.4",
            # GroQ
            "groq==0.11.0",
            "langchain-groq==0.1.4",
            # HF
            "llama-index-llms-huggingface==0.2.7",
            # Google
            "langchain-google-vertexai==1.0.8",
            "langchain-google-genai==1.0.8",
            "google-generativeai==0.7.2",
            "vertexai==1.60.0",
            "google-cloud-aiplatform>=1.60.0",
            "grpc-google-iam-v1==0.13.0",
            # OpenAI
            "langchain-openai==0.1.21",
            "openai==1.40.8",
            "llama-index-llms-openai==0.1.11",
            # Antropic
            "langchain-anthropic==0.1.23",
            "anthropic==0.34.0",
        ]
    },
    tests_require=[
        'pytest>=7.2.2',
        'pytest-asyncio==0.21.1',
        'pytest-xdist==3.3.1',
        'pytest-assume==2.4.3'
    ],
    test_suite='tests',
    ext_modules=cythonize(extensions),
    # cmdclass={"build_ext": BuildExtensions},
    project_urls={  # Optional
        'Source': 'https://github.com/phenobarbital/ai-parrot',
        'Tracker': 'https://github.com/phenobarbital/ai-parrot/issues',
        'Documentation': 'https://github.com/phenobarbital/ai-parrot/',
        "Funding": "https://paypal.me/phenobarbital",
        "Say Thanks!": "https://saythanks.io/to/phenobarbital",
    },
)



from collections.abc import Callable
from pathlib import PurePath
from typing import Any
import re
from langchain_community.document_loaders import ReadTheDocsLoader as RTLoader
from .abstract import AbstractLoader


class ReadTheDocsLoader(AbstractLoader):
    """
    Loading documents from ReadTheDocs.
    """
    _extension: list = []

    def __init__(
        self,
        path: PurePath,
        tokenizer: Callable[..., Any] = None,
        text_splitter: Callable[..., Any] = None,
        source_type: str = 'documentation',
        encoding: str = 'utf-8',
        origin: str = '',
        **kwargs
    ):
        super().__init__(tokenizer, text_splitter, source_type=source_type, **kwargs)
        self.path = path
        self.encoding = encoding
        self.origin = origin
        self._prefix = ''
        match = re.search(r'://([^/]+)', origin)
        if match:
            self._prefix = match.group(1)

    def load(self) -> list:
        documents = []
        self.logger.info(
            f"Loading ReadTheDocs from: {self.path}"
        )
        rt_loader = RTLoader(path=self.path, encoding=self.encoding)
        documents = rt_loader.load()
        for doc in documents:
            src = doc.metadata.get('source')
            suffix = src.replace(f'{self.path}', '').replace(self._prefix, '')
            if suffix.startswith('//'):
                suffix = suffix[1:]
            url = f"{self.origin}{suffix}"
            metadata = {
                "source": url,
                "url": url,
                "index": suffix,
                "filename": src,
                "question": '',
                "answer": '',
                'type': 'documentation',
                "source_type": self._source_type,
                "summary": '',
                "document_meta": {
                    **doc.metadata
                }
            }
            doc.metadata = metadata
        return documents

    def parse(self, source):
        pass



from typing import Any
import os
from navconfig import config
from langchain_community.tools.asknews import AskNewsSearch
from langchain.tools import BaseTool

class AskNewsTool(BaseTool):
    """Tool that searches the AskNews API."""
    name: str = "asknews_search"
    description: str = (
        "This tool allows you to perform a search on up-to-date news and historical "
        "news. If you needs news from more than 48 hours ago, you can estimate the "
        "number of hours back to search."
    )
    search: Any = None

    def __init__(self, max_results: int = 5, **kwargs):
        super().__init__(**kwargs)
        os.environ["ASKNEWS_CLIENT_ID"] = config.get('ASKNEWS_CLIENT_ID')
        os.environ["ASKNEWS_CLIENT_SECRET"] = config.get('ASKNEWS_CLIENT_SECRET')
        self.search = AskNewsSearch(max_results=5)

    def _run(
        self,
        query: str
    ) -> str:
        """Use the Wikipedia tool."""
        return self.search.invoke(
            {
                "query": query,
            }
        )



from pathlib import PurePath
from langchain_core.document_loaders.blob_loaders import Blob
from langchain_community.document_loaders.generic import GenericLoader
from langchain_community.document_loaders.parsers import LanguageParser
from langchain_community.document_loaders import (
    DirectoryLoader,
    TextLoader,
    JSONLoader
)
from langchain_text_splitters import Language
from langchain.text_splitter import (
    RecursiveCharacterTextSplitter
)
from .abstract import AbstractLoader


class RepositoryLoader(AbstractLoader):
    """Repository (Code Directory) loader.
    """
    exclude_paths: list = [
        ".venv/**",
        ".venv/**/**/*",
        ".git/**",
        "node_modules/**",
        "build/**",
        "dist/**",
        "templates/**",
        "tmp/**"
    ]

    def load(self, path: PurePath, lang: str = 'python', excludes: list = []) -> list:
        """
        Load data from a repository and return it as a Langchain Document.
        """
        if isinstance(path, str):
            path = PurePath(path)
        if excludes:
            self.exclude_paths += excludes
        excludes_path = [
            str(path.joinpath(p).resolve()) for p in self.exclude_paths
        ]
        if lang == 'python':
            parser = LanguageParser(language=Language.PYTHON, parser_threshold=100)
            splitter = RecursiveCharacterTextSplitter.from_language(
                language=Language.PYTHON, chunk_size=1024, chunk_overlap=200
            )
            suffixes = [".py", ".pyx"]
            glob = "**/[!.]*.py?"
        elif lang == 'javascript':
            parser = LanguageParser(language=Language.JS, parser_threshold=100)
            splitter = RecursiveCharacterTextSplitter.from_language(
                language=Language.JS, chunk_size=1024, chunk_overlap=200
            )
            suffixes = [".js", ".jsx", ".json", ".ts", ".tsx"]
        elif lang == 'typescript':
            parser = LanguageParser(language=Language.TS, parser_threshold=100)
            splitter = RecursiveCharacterTextSplitter.from_language(
                language=Language.TS, chunk_size=1024, chunk_overlap=200
            )
            suffixes = [".js", ".jsx", ".json", ".ts", ".tsx"]
        elif lang == 'json':
            loader = DirectoryLoader(
                path,
                glob="**/*.json",
                show_progress=True,
                exclude=excludes_path,
                silent_errors=True,
                recursive=True,
                # loader_cls=TextLoader,
                loader_cls=JSONLoader,
                loader_kwargs={
                    'jq_schema': '.',
                    'text_content': False
                }
            )
            docs = loader.load()
            for doc in docs:
                doc.metadata['url'] = ''
                doc.metadata['source_type'] = self._source_type
                doc.metadata['language'] = lang
            return self.text_splitter.split_documents(docs)
        else:
            raise ValueError(
                f"Language {lang} not supported for Repository"
            )
        loader = GenericLoader.from_filesystem(
            path,
            glob=glob,
            suffixes=suffixes,
            exclude=self.exclude_paths,
            parser=parser,
            show_progress=True
        )
        docs = loader.load()
        for doc in docs:
            doc.metadata['url'] = ''
            doc.metadata['source_type'] = self._source_type
            doc.metadata['language'] = lang
        documents = splitter.split_documents(docs)
        return documents

    def parse(self, source):
        raise NotImplementedError("Parser method is not implemented for PDFLoader.")



from collections.abc import Callable
from typing import Any, Optional, List
from pathlib import Path, PurePath
from io import StringIO
import fitz
import pandas as pd
from langchain.docstore.document import Document
from .basepdf import BasePDF


class PDFTablesLoader(BasePDF):
    """
    Loader for Tables in PDF Files.
    """
    _extension = ['.pdf']

    def __init__(
        self,
        path: PurePath,
        tokenizer: Callable[..., Any] = None,
        text_splitter: Callable[..., Any] = None,
        source_type: str = 'pdf',
        language: str = "eng",
        table_settings: dict = {},
        **kwargs
    ):
        super().__init__(
            path,
            tokenizer,
            text_splitter,
            source_type,
            language=language,
            **kwargs
        )
        # Table Settings:
        self.table_settings = {
            #"vertical_strategy": "text",
            # "horizontal_strategy": "text",
            "intersection_x_tolerance": 5,
            "intersection_y_tolerance": 5
        }
        if table_settings:
            self.table_settings.update(table_settings)
        self._skiprows = kwargs.pop('skiprows', None)

    def unique_columns(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Rename duplicate columns in the DataFrame to ensure they are unique.

        Args:
            df (pd.DataFrame): The DataFrame with potential duplicate column names.

        Returns:
            pd.DataFrame: A DataFrame with unique column names.
        """
        seen = {}
        new_columns = []
        for col in df.columns:
            new_col = col
            count = seen.get(col, 0)
            while new_col in new_columns:
                count += 1
                new_col = f"{col}_{count}"
            new_columns.append(new_col)
            seen[col] = count
        df.columns = new_columns
        return df

    def get_markdown(self, df: pd.DataFrame) -> str:
        """
        Convert a DataFrame to a Markdown string.

        Args:
            df (pd.DataFrame): The DataFrame to convert.

        Returns:
            str: The JSON string.
        """
        buffer = StringIO()
        df = self.unique_columns(df)
        df.to_markdown(buffer)
        buffer.seek(0)
        return buffer.getvalue()

    def parse_table(self, table_idx, table, page_number, path) -> pd.DataFrame:
        df = table.to_pandas()  # convert to pandas DataFrame
        df = df.dropna(axis=1, how='all')
        df = df.dropna(how='all', axis=0)  # Drop empty rows
        page = page_number + 1
        table_meta = {
            "url": '',
            "source": f"{path.name} Page.#{page} Table.#{table_idx}",
            "filename": path.name,
            "index": f"{path.name}:Table:{table_idx}",
            "question": '',
            "answer": '',
            "type": 'table',
            "data": {},
            "summary": '',
            "document_meta": {
                "table_index": table_idx,
                "table_shape": df.shape,
                "table_columns": df.columns.tolist(),
                "description": f"Extracted from Page.#{page}."
            },
            "source_type": self._source_type
        }
        return df, table_meta

    def _load_pdf(self, path: Path) -> list:
        """
        Load a PDF file using the Fitz library.

        Args:
            path (Path): The path to the PDF file.

        Returns:
            list: A list of Langchain Documents.
        """
        if self._check_path(path):
            self.logger.info(f"Loading PDF file: {path}")
            pdf = fitz.open(str(path))  # Open the PDF file
            docs = []
            for page_number in range(pdf.page_count):
                page = pdf[page_number]
                try:
                    tabs = page.find_tables(**self.table_settings)
                    for tab_idx, tab in enumerate(tabs):
                        df, _meta = self.parse_table(tab_idx, tab, page_number, path)
                        ## Sample information:
                        print('::: Printing Table Information === ')
                        print(df)
                        print("::: Printing Column Information === ")
                        for column, t in df.dtypes.items():
                            print(column, "->", t, "->", df[column].iloc[0])
                        # convert into markdown:
                        txt = df.to_markdown()
                        if txt:
                            docs.append(
                                Document(page_content=txt, metadata=_meta)
                            )
                except Exception as exc:
                    print(exc)
                    continue
            return docs



import asyncio
import aiohttp
from markdownify import MarkdownConverter
from bs4 import BeautifulSoup as bs
from langchain.text_splitter import MarkdownTextSplitter
from langchain.docstore.document import Document
from .abstract import AbstractLoader


class WebBaseLoader(AbstractLoader):
    """Class to load web pages and extract text as Markdown."""
    def __init__(self, urls: dict, source_type: str = 'website', **kwargs):
        self.urls = urls
        self._source_type = source_type
        self.timeout: int = kwargs.pop('timeout', 60)
        self._wait: int = kwargs.pop('wait', 60)
        super().__init__(source_type=source_type, **kwargs)
        self.md_splitter = MarkdownTextSplitter(
            chunk_size = 1024,
            chunk_overlap=10
        )

    def md(self, soup, **options):
        return MarkdownConverter(**options).convert_soup(soup)

    async def _fetch(
        self, url: str, retries: int = 3, cooldown: int = 2, backoff: float = 1.5
    ) -> str:
        async with aiohttp.ClientSession() as session:
            for i in range(retries):
                try:
                    async with session.get(url, allow_redirects=True) as response:
                        if response.status == 200:
                            return await response.text()
                        elif response.status == 429:
                            # Too many requests
                            if i == retries - 1:
                                return ''
                            self.logger.warning(
                                f"Too many requests to {url}. Waiting {self._wait} seconds."
                            )
                            await asyncio.sleep(self._wait)
                        else:  # Other non-success status codes
                            response.raise_for_status()  # Raise for other errors
                except aiohttp.ClientConnectionError as e:
                    if i == retries - 1:
                        raise
                    else:
                        self.logger.warning(
                            f"Error fetching {url} with attempt "
                            f"{i + 1}/{retries}: {e}. Retrying..."
                        )
                        await asyncio.sleep(cooldown * backoff**i)
                except aiohttp.ClientResponseError as e:
                    self.logger.warning(
                        f"Request failed (ClientResponseError): {e}"
                    )
                    return ''
        self.logger.warning(f"Failed to fetch {url} after {retries} attempts.")
        return ''

    async def get_address(self, url: str) -> list:
        self.logger.info(
            f'Downloading URL {url}'
        )
        html = await self._fetch(url)
        docs = []
        if html:
            soup = bs(html, 'html.parser')
            md_text = self.md(soup)
            try:
                title = soup.title.string
            except AttributeError:
                title = None
            metadata = {
                "url": url,
                "source": url,
                "index": "",
                "filename": '',
                "type": 'webpage',
                "question": '',
                "answer": '',
                "source_type": self._source_type,
                "summary": "",
                "document_meta": {
                    "title": title,
                    "description": soup.find('meta', attrs={'name': 'description'})['content']
                    if soup.find('meta', attrs={'name': 'description'}) else '',
                    "keywords": soup.find('meta', attrs={'name': 'keywords'})['content']
                    if soup.find('meta', attrs={'name': 'keywords'}) else '',
                }
            }
            for chunk in self.md_splitter.split_text(md_text):
                docs.append(
                    Document(
                        page_content=chunk,
                        metadata=metadata
                    )
                )
        return docs

    async def load(self, **kwargs) -> list:
        documents = []
        if self.urls is None:
            return []
        for address in self.urls:
            docs = await self.get_address(address)
            documents.extend(docs)
        return self.split_documents(documents)

    def parse(self, source):
        pass



from langchain_community.tools.yahoo_finance_news import YahooFinanceNewsTool
from langchain_community.tools import YouTubeSearchTool
from langchain_community.agent_toolkits import O365Toolkit
from navconfig import config
from .wikipedia import WikipediaTool, WikidataTool
from .asknews import AskNewsTool
from .duck import DuckDuckGoSearchTool, DuckDuckGoRelevantSearch
from .weather import OpenWeather, OpenWeatherMapTool
from .google import GoogleLocationFinder, GoogleSiteSearchTool, GoogleSearchTool
from .zipcode import ZipcodeAPIToolkit
from .bing import BingSearchTool
from .stack import StackExchangeTool



"""
A Directory for adding more Tools to CrewAI
"""
from .duckgo import DuckDuckGoRelevantSearch, DuckDuckGoSearchTool
from .rag import RagSearchTool
from .file import SaveFile
from .bing import BingSearchTool
from .md2pdf import MarkdownToPDFTool
from .google import GoogleSearchTool, GoogleSiteSearchTool, GoogleLocationFinder


__all__ = [
    'GoogleSearchTool',
    'GoogleSiteSearchTool',
    'BingSearchTool',
    'DuckDuckGoRelevantSearch',
    'DuckDuckGoSearchTool',
    'RagSearchTool',
    'SaveFile',
    'MarkdownToPDFTool',
    'GoogleLocationFinder',
]



from .dir import load_directory
from .pdf import PDFLoader
from .web import WebLoader
from .youtube import YoutubeLoader
from .vimeo import VimeoLoader
from .word import MSWordLoader
from .ppt import PPTXLoader
from .repo import RepositoryLoader
from .github import GithubLoader
from .json import JSONLoader
from .excel import ExcelLoader
from .web_base import WebBaseLoader
from .pdfmark import PDFMarkdownLoader
from .pdfimages import PDFImageLoader
from .pdftables import PDFTablesLoader
from .pdfchapters import PDFChapterLoader
from .txt import TXTLoader
from .qa import QAFileLoader
from .rtd import ReadTheDocsLoader
from .videolocal import VideoLocalLoader
